from collections import OrderedDict, defaultdict
from collections.abc import Iterable, Iterator, Mapping, Sequence, Set
from typing import (  # noqa: UP035
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Callable,
    Optional,
    TypeVar,
    Union,
    cast,
)

from toposort import CircularDependencyError
from typing_extensions import Self

import dagster._check as check
from dagster._annotations import deprecated_param, public
from dagster._core.definitions.config import ConfigMapping
from dagster._core.definitions.definition_config_schema import IDefinitionConfigSchema
from dagster._core.definitions.dependency import (
    DependencyMapping,
    DependencyStructure,
    GraphNode,
    Node,
    NodeHandle,
    NodeInput,
    NodeInputHandle,
    NodeInvocation,
    NodeOutput,
    NodeOutputHandle,
)
from dagster._core.definitions.hook_definition import HookDefinition
from dagster._core.definitions.input import (
    FanInInputPointer,
    InputDefinition,
    InputMapping,
    InputPointer,
)
from dagster._core.definitions.logger_definition import LoggerDefinition
from dagster._core.definitions.metadata import RawMetadataValue
from dagster._core.definitions.node_container import (
    create_execution_structure,
    normalize_dependency_dict,
)
from dagster._core.definitions.node_definition import NodeDefinition
from dagster._core.definitions.output import OutputDefinition, OutputMapping
from dagster._core.definitions.policy import RetryPolicy
from dagster._core.definitions.resource_requirement import ResourceRequirement
from dagster._core.errors import DagsterInvalidDefinitionError, DagsterInvariantViolationError
from dagster._core.selector.subset_selector import AssetSelectionData
from dagster._core.types.dagster_type import (
    DagsterType,
    DagsterTypeKind,
    construct_dagster_type_dictionary,
)
from dagster._core.utils import toposort_flatten
from dagster._utils.warnings import normalize_renamed_param

if TYPE_CHECKING:
    from dagster._core.definitions.asset_layer import AssetLayer
    from dagster._core.definitions.assets import AssetsDefinition
    from dagster._core.definitions.composition import PendingNodeInvocation
    from dagster._core.definitions.executor_definition import ExecutorDefinition
    from dagster._core.definitions.job_definition import JobDefinition
    from dagster._core.definitions.op_definition import OpDefinition
    from dagster._core.definitions.partition import PartitionedConfig, PartitionsDefinition
    from dagster._core.definitions.run_config import RunConfig
    from dagster._core.definitions.source_asset import SourceAsset
    from dagster._core.execution.execute_in_process_result import ExecuteInProcessResult
    from dagster._core.instance import DagsterInstance

T = TypeVar("T")


def _check_node_defs_arg(
    graph_name: str, node_defs: Optional[Sequence[NodeDefinition]]
) -> Sequence[NodeDefinition]:
    node_defs = node_defs or []

    _node_defs = check.opt_sequence_param(node_defs, "node_defs")
    for node_def in _node_defs:
        if isinstance(node_def, NodeDefinition):
            continue
        elif callable(node_def):
            raise DagsterInvalidDefinitionError(
                f"""You have passed a lambda or function {node_def.__name__} into {graph_name} that is
                not a node. You have likely forgetten to annotate this function with
                the @op or @graph decorators.'
                """
            )
        else:
            raise DagsterInvalidDefinitionError(f"Invalid item in node list: {node_def!r}")

    return node_defs


def create_adjacency_lists(
    nodes: Sequence[Node],
    dep_structure: DependencyStructure,
) -> tuple[Mapping[str, set[str]], Mapping[str, set[str]]]:
    visit_dict = {s.name: False for s in nodes}
    forward_edges: dict[str, set[str]] = {s.name: set() for s in nodes}
    backward_edges: dict[str, set[str]] = {s.name: set() for s in nodes}

    def visit(node_name: str) -> None:
        if visit_dict[node_name]:
            return

        visit_dict[node_name] = True

        for node_output in dep_structure.all_upstream_outputs_from_node(node_name):
            forward_node = node_output.node.name
            backward_node = node_name
            if forward_node in forward_edges:
                forward_edges[forward_node].add(backward_node)
                backward_edges[backward_node].add(forward_node)
                visit(forward_node)

    for s in nodes:
        visit(s.name)

    return (forward_edges, backward_edges)


@deprecated_param(
    param="node_input_source_assets",
    breaking_version="2.0",
    additional_warn_text="Use `input_assets` instead.",
)
class GraphDefinition(NodeDefinition):
    """Defines a Dagster op graph.

    An op graph is made up of

    - Nodes, which can either be an op (the functional unit of computation), or another graph.
    - Dependencies, which determine how the values produced by nodes as outputs flow from
      one node to another. This tells Dagster how to arrange nodes into a directed, acyclic graph
      (DAG) of compute.

    End users should prefer the :func:`@graph <graph>` decorator. GraphDefinition is generally
    intended to be used by framework authors or for programatically generated graphs.

    Args:
        name (str): The name of the graph. Must be unique within any :py:class:`GraphDefinition`
            or :py:class:`JobDefinition` containing the graph.
        description (Optional[str]): A human-readable description of the job.
        node_defs (Optional[Sequence[NodeDefinition]]): The set of ops / graphs used in this graph.
        dependencies (Optional[Dict[Union[str, NodeInvocation], Dict[str, DependencyDefinition]]]):
            A structure that declares the dependencies of each op's inputs on the outputs of other
            ops in the graph. Keys of the top level dict are either the string names of ops in the
            graph or, in the case of aliased ops, :py:class:`NodeInvocations <NodeInvocation>`.
            Values of the top level dict are themselves dicts, which map input names belonging to
            the op or aliased op to :py:class:`DependencyDefinitions <DependencyDefinition>`.
        input_mappings (Optional[Sequence[InputMapping]]): Defines the inputs to the nested graph, and
            how they map to the inputs of its constituent ops.
        output_mappings (Optional[Sequence[OutputMapping]]): Defines the outputs of the nested graph,
            and how they map from the outputs of its constituent ops.
        config (Optional[ConfigMapping]): Defines the config of the graph, and how its schema maps
            to the config of its constituent ops.
        tags (Optional[Dict[str, Any]]): Arbitrary metadata for any execution of the graph.
            Values that are not strings will be json encoded and must meet the criteria that
            `json.loads(json.dumps(value)) == value`.  These tag values may be overwritten by tag
            values provided at invocation time.
        composition_fn (Optional[Callable]): The function that defines this graph. Used to generate
            code references for this graph.

    Examples:
        .. code-block:: python

            @op
            def return_one():
                return 1

            @op
            def add_one(num):
                return num + 1

            graph_def = GraphDefinition(
                name='basic',
                node_defs=[return_one, add_one],
                dependencies={'add_one': {'num': DependencyDefinition('return_one')}},
            )
    """

    _node_defs: Sequence[NodeDefinition]
    _dagster_type_dict: Mapping[str, DagsterType]
    _dependencies: DependencyMapping[NodeInvocation]
    _dependency_structure: DependencyStructure
    _node_dict: Mapping[str, Node]
    _input_mappings: Sequence[InputMapping]
    _output_mappings: Sequence[OutputMapping]
    _config_mapping: Optional[ConfigMapping]
    _nodes_in_topological_order: Sequence[Node]

    # (node name within the graph -> (input name -> AssetsDefinition to load that input from))
    # Does NOT include keys for:
    # - Inputs to the graph itself
    # - Inputs to nodes within sub-graphs of the graph
    _input_assets: Mapping[str, Mapping[str, "AssetsDefinition"]]

    def __init__(
        self,
        name: str,
        *,
        description: Optional[str] = None,
        node_defs: Optional[Sequence[NodeDefinition]] = None,
        dependencies: Optional[
            Union[DependencyMapping[str], DependencyMapping[NodeInvocation]]
        ] = None,
        input_mappings: Optional[Sequence[InputMapping]] = None,
        output_mappings: Optional[Sequence[OutputMapping]] = None,
        config: Optional[ConfigMapping] = None,
        tags: Optional[Mapping[str, str]] = None,
        node_input_source_assets: Optional[Mapping[str, Mapping[str, "SourceAsset"]]] = None,
        input_assets: Optional[
            Mapping[str, Mapping[str, Union["AssetsDefinition", "SourceAsset"]]]
        ] = None,
        composition_fn: Optional[Callable] = None,
        **kwargs: Any,
    ):
        from dagster._core.definitions.external_asset import create_external_asset_from_source_asset
        from dagster._core.definitions.source_asset import SourceAsset

        self._node_defs = _check_node_defs_arg(name, node_defs)

        # `dependencies` will be converted to `dependency_structure` and `node_dict`, which may
        # alternatively be passed directly (useful when copying)
        self._dependencies = normalize_dependency_dict(dependencies)
        self._dependency_structure, self._node_dict = create_execution_structure(
            self._node_defs, self._dependencies, graph_definition=self
        )

        # Sequence[InputMapping]
        self._input_mappings = check.opt_sequence_param(input_mappings, "input_mappings")
        input_defs = _validate_in_mappings(
            self._input_mappings,
            self._node_dict,
            self._dependency_structure,
            name,
            class_name=type(self).__name__,
        )

        # Sequence[OutputMapping]
        self._output_mappings, output_defs = _validate_out_mappings(
            check.opt_sequence_param(output_mappings, "output_mappings"),
            self._node_dict,
            name,
            class_name=type(self).__name__,
        )

        self._config_mapping = check.opt_inst_param(config, "config", ConfigMapping)

        self._composition_fn = check.opt_callable_param(composition_fn, "composition_fn")

        super().__init__(
            name=name,
            description=description,
            input_defs=input_defs,
            output_defs=output_defs,
            tags=tags,
            **kwargs,
        )

        # must happen after base class construction as properties are assumed to be there
        # eager computation to detect cycles
        self._nodes_in_topological_order = self._get_nodes_in_topological_order()
        self._dagster_type_dict = construct_dagster_type_dictionary([self])

        # Backcompat: the previous  API `node_input_source_assets` with a Dict[str, Dict[str,
        # SourceAsset]]. The new API is `input_assets` and accepts external assets as well as
        # SourceAsset.
        self._input_assets = {}
        input_assets = check.opt_mapping_param(
            normalize_renamed_param(
                new_val=input_assets,
                new_arg="input_assets",
                old_val=node_input_source_assets,
                old_arg="node_input_source_assets",
            ),
            "input_assets",
            key_type=str,
            value_type=dict,
        )
        for node_name, inputs in input_assets.items():
            self._input_assets[node_name] = {
                input_name: (
                    create_external_asset_from_source_asset(asset)
                    if isinstance(asset, SourceAsset)
                    else asset
                )
                for input_name, asset in inputs.items()
            }

    def _get_nodes_in_topological_order(self) -> Sequence[Node]:
        _forward_edges, backward_edges = create_adjacency_lists(
            self.nodes, self.dependency_structure
        )

        try:
            order = toposort_flatten(backward_edges)
        except CircularDependencyError as err:
            raise DagsterInvalidDefinitionError(str(err)) from err

        return [self.node_named(node_name) for node_name in order]

    def get_inputs_must_be_resolved_top_level(
        self, asset_layer: "AssetLayer", handle: Optional[NodeHandle] = None
    ) -> Sequence[InputDefinition]:
        unresolveable_input_defs: list[InputDefinition] = []
        for node in self.node_dict.values():
            cur_handle = NodeHandle(node.name, handle)
            for input_def in node.definition.get_inputs_must_be_resolved_top_level(
                asset_layer, cur_handle
            ):
                if self.dependency_structure.has_deps(NodeInput(node, input_def)):
                    continue
                elif not node.container_maps_input(input_def.name):
                    raise DagsterInvalidDefinitionError(
                        f"Input '{input_def.name}' of {node.describe_node()} "
                        "has no way of being resolved. Must provide a resolution to this "
                        "input via another op/graph, or via a direct input value mapped from the "
                        "top-level graph. To "
                        "learn more, see the docs for unconnected inputs: "
                        "https://legacy-docs.dagster.io/concepts/io-management/unconnected-inputs#unconnected-inputs."
                    )
                else:
                    mapped_input = node.container_mapped_input(input_def.name)
                    unresolveable_input_defs.append(mapped_input.get_definition())
        return unresolveable_input_defs

    @property
    def node_type_str(self) -> str:
        return "graph"

    @property
    def is_graph_job_op_node(self) -> bool:
        return True

    @property
    def nodes(self) -> Sequence[Node]:
        return list(set(self._node_dict.values()))

    @property
    def node_dict(self) -> Mapping[str, Node]:
        return self._node_dict

    @property
    def node_defs(self) -> Sequence[NodeDefinition]:
        return self._node_defs

    @property
    def nodes_in_topological_order(self) -> Sequence[Node]:
        return self._nodes_in_topological_order

    @property
    def input_assets(self) -> Mapping[str, Mapping[str, "AssetsDefinition"]]:
        return self._input_assets

    @property
    def composition_fn(self) -> Optional[Callable]:
        return self._composition_fn

    def has_node_named(self, name: str) -> bool:
        check.str_param(name, "name")
        return name in self._node_dict

    def node_named(self, name: str) -> Node:
        check.str_param(name, "name")
        if name not in self._node_dict:
            raise DagsterInvariantViolationError(f"{self._name} has no op named {name}.")

        return self._node_dict[name]

    def get_node(self, handle: NodeHandle) -> Node:
        check.inst_param(handle, "handle", NodeHandle)
        current = handle
        lineage: list[str] = []
        while current:
            lineage.append(current.name)
            current = current.parent

        name = lineage.pop()
        node = self.node_named(name)
        while lineage:
            name = lineage.pop()
            # We know that this is a current node is a graph while ascending lineage
            definition = cast(GraphDefinition, node.definition)
            node = definition.node_named(name)

        return node

    def iterate_node_defs(self) -> Iterator[NodeDefinition]:
        yield self
        for outer_node_def in self._node_defs:
            yield from outer_node_def.iterate_node_defs()

    def iterate_op_defs(self) -> Iterator["OpDefinition"]:
        for outer_node_def in self._node_defs:
            yield from outer_node_def.iterate_op_defs()

    def iterate_node_handles(
        self, parent_node_handle: Optional[NodeHandle] = None
    ) -> Iterator[NodeHandle]:
        for node in self.node_dict.values():
            cur_node_handle = NodeHandle(node.name, parent_node_handle)
            if isinstance(node, GraphNode):
                yield from node.definition.iterate_node_handles(cur_node_handle)
            yield cur_node_handle

    @public
    @property
    def input_mappings(self) -> Sequence[InputMapping]:
        """Input mappings for the graph.

        An input mapping is a mapping from an input of the graph to an input of a child node.
        """
        return self._input_mappings

    @public
    @property
    def output_mappings(self) -> Sequence[OutputMapping]:
        """Output mappings for the graph.

        An output mapping is a mapping from an output of the graph to an output of a child node.
        """
        return self._output_mappings

    @public
    @property
    def config_mapping(self) -> Optional[ConfigMapping]:
        """The config mapping for the graph, if present.

        By specifying a config mapping function, you can override the configuration for the child nodes contained within a graph.
        """
        return self._config_mapping

    @property
    def has_config_mapping(self) -> bool:
        return self._config_mapping is not None

    def all_dagster_types(self) -> Iterable[DagsterType]:
        return self._dagster_type_dict.values()

    def has_dagster_type(self, name: str) -> bool:
        check.str_param(name, "name")
        return name in self._dagster_type_dict

    def dagster_type_named(self, name: str) -> DagsterType:
        check.str_param(name, "name")
        return self._dagster_type_dict[name]

    def get_input_mapping(self, input_name: str) -> InputMapping:
        check.str_param(input_name, "input_name")
        for mapping in self._input_mappings:
            if mapping.graph_input_name == input_name:
                return mapping
        check.failed(f"Could not find input mapping {input_name}")

    def input_mapping_for_pointer(
        self, pointer: Union[InputPointer, FanInInputPointer]
    ) -> Optional[InputMapping]:
        check.inst_param(pointer, "pointer", (InputPointer, FanInInputPointer))

        for mapping in self._input_mappings:
            if mapping.maps_to == pointer:
                return mapping
        return None

    def get_output_mapping(self, output_name: str) -> OutputMapping:
        check.str_param(output_name, "output_name")
        for mapping in self._output_mappings:
            if mapping.graph_output_name == output_name:
                return mapping
        check.failed(f"Could not find output mapping {output_name}")

    def resolve_output_to_origin(
        self, output_name: str, handle: Optional[NodeHandle]
    ) -> tuple[OutputDefinition, Optional[NodeHandle]]:
        check.str_param(output_name, "output_name")
        check.opt_inst_param(handle, "handle", NodeHandle)

        mapping = self.get_output_mapping(output_name)
        check.invariant(mapping, "Can only resolve outputs for valid output names")
        mapped_node = self.node_named(mapping.maps_from.node_name)
        return mapped_node.definition.resolve_output_to_origin(
            mapping.maps_from.output_name,
            NodeHandle(mapped_node.name, handle),
        )

    def resolve_output_to_origin_op_def(self, output_name: str) -> "OpDefinition":
        mapping = self.get_output_mapping(output_name)
        check.invariant(mapping, "Can only resolve outputs for valid output names")
        return self.node_named(
            mapping.maps_from.node_name
        ).definition.resolve_output_to_origin_op_def(output_name)

    def default_value_for_input(self, input_name: str) -> object:
        check.str_param(input_name, "input_name")

        # base case
        if self.input_def_named(input_name).has_default_value:
            return self.input_def_named(input_name).default_value

        mapping = self.get_input_mapping(input_name)
        check.invariant(mapping, "Can only resolve inputs for valid input names")
        mapped_node = self.node_named(mapping.maps_to.node_name)

        return mapped_node.definition.default_value_for_input(mapping.maps_to.input_name)

    def input_has_default(self, input_name: str) -> bool:
        check.str_param(input_name, "input_name")

        # base case
        if self.input_def_named(input_name).has_default_value:
            return True

        mapping = self.get_input_mapping(input_name)
        check.invariant(mapping, "Can only resolve inputs for valid input names")
        mapped_node = self.node_named(mapping.maps_to.node_name)

        return mapped_node.definition.input_has_default(mapping.maps_to.input_name)

    @property
    def dependencies(self) -> DependencyMapping[NodeInvocation]:
        return self._dependencies

    @property
    def dependency_structure(self) -> DependencyStructure:
        return self._dependency_structure

    @property
    def config_schema(self) -> Optional[IDefinitionConfigSchema]:
        return self.config_mapping.config_schema if self.config_mapping is not None else None

    def input_supports_dynamic_output_dep(self, input_name: str) -> bool:
        mapping = self.get_input_mapping(input_name)
        target_node = mapping.maps_to.node_name
        # check if input mapped to node which is downstream of another dynamic output within
        if self.dependency_structure.is_dynamic_mapped(target_node):
            return False

        # check if input mapped to node which starts new dynamic downstream
        if self.dependency_structure.has_dynamic_downstreams(target_node):
            return False

        return self.node_named(target_node).definition.input_supports_dynamic_output_dep(
            mapping.maps_to.input_name
        )

    def copy(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        input_mappings: Optional[Sequence[InputMapping]] = None,
        output_mappings: Optional[Sequence[OutputMapping]] = None,
        config: Optional[ConfigMapping] = None,
        tags: Optional[Mapping[str, str]] = None,
        input_assets: Optional[Mapping[str, Mapping[str, "AssetsDefinition"]]] = None,
    ) -> Self:
        return self.__class__(
            node_defs=self.node_defs,
            dependencies=self.dependencies,
            name=name or self.name,
            description=description or self.description,
            input_mappings=input_mappings or self._input_mappings,
            output_mappings=output_mappings or self._output_mappings,
            config=config or self.config_mapping,
            tags=tags or self.tags,
            input_assets=input_assets or self._input_assets,
        )

    def copy_for_configured(
        self,
        name: str,
        description: Optional[str],
        config_schema: Any,
    ) -> Self:
        if not self.has_config_mapping:
            raise DagsterInvalidDefinitionError(
                "Only graphs utilizing config mapping can be pre-configured. The graph "
                f'"{self.name}" does not have a config mapping, and thus has nothing to be '
                "configured."
            )
        config_mapping = cast(ConfigMapping, self.config_mapping)
        return self.copy(
            name=name,
            description=check.opt_str_param(description, "description", default=self.description),
            config=ConfigMapping(
                config_mapping.config_fn,
                config_schema=config_schema,
                receive_processed_config_values=config_mapping.receive_processed_config_values,
            ),
        )

    def node_names(self) -> Sequence[str]:
        return list(self._node_dict.keys())

    @public
    def to_job(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        resource_defs: Optional[Mapping[str, object]] = None,
        config: Optional[
            Union["RunConfig", ConfigMapping, Mapping[str, object], "PartitionedConfig"]
        ] = None,
        tags: Optional[Mapping[str, object]] = None,
        metadata: Optional[Mapping[str, RawMetadataValue]] = None,
        logger_defs: Optional[Mapping[str, LoggerDefinition]] = None,
        executor_def: Optional["ExecutorDefinition"] = None,
        hooks: Optional[AbstractSet[HookDefinition]] = None,
        op_retry_policy: Optional[RetryPolicy] = None,
        op_selection: Optional[Sequence[str]] = None,
        partitions_def: Optional["PartitionsDefinition"] = None,
        asset_layer: Optional["AssetLayer"] = None,
        input_values: Optional[Mapping[str, object]] = None,
        run_tags: Optional[Mapping[str, object]] = None,
        _asset_selection_data: Optional[AssetSelectionData] = None,
    ) -> "JobDefinition":
        """Make this graph in to an executable Job by providing remaining components required for execution.

        Args:
            name (Optional[str]):
                The name for the Job. Defaults to the name of the this graph.
            resource_defs (Optional[Mapping [str, object]]):
                Resources that are required by this graph for execution.
                If not defined, `io_manager` will default to filesystem.
            config:
                Describes how the job is parameterized at runtime.

                If no value is provided, then the schema for the job's run config is a standard
                format based on its ops and resources.

                If a dictionary is provided, then it must conform to the standard config schema, and
                it will be used as the job's run config for the job whenever the job is executed.
                The values provided will be viewable and editable in the Dagster UI, so be
                careful with secrets.

                If a :py:class:`ConfigMapping` object is provided, then the schema for the job's run config is
                determined by the config mapping, and the ConfigMapping, which should return
                configuration in the standard format to configure the job.

                If a :py:class:`PartitionedConfig` object is provided, then it defines a discrete set of config
                values that can parameterize the job, as well as a function for mapping those
                values to the base config. The values provided will be viewable and editable in the
                Dagster UI, so be careful with secrets.
            tags (Optional[Mapping[str, object]]): A set of key-value tags that annotate the job and can
                be used for searching and filtering in the UI. Values that are not already strings will
                be serialized as JSON. If `run_tags` is not set, then the content of `tags` will also be
                automatically appended to the tags of any runs of this job.
            run_tags (Optional[Mapping[str, object]]):
                A set of key-value tags that will be automatically attached to runs launched by this
                job. Values that are not already strings will be serialized as JSON. These tag values
                may be overwritten by tag values provided at invocation time. If `run_tags` is set, then
                `tags` are not automatically appended to the tags of any runs of this job.
            metadata (Optional[Mapping[str, RawMetadataValue]]):
                Arbitrary information that will be attached to the JobDefinition and be viewable in the Dagster UI.
                Keys must be strings, and values must be python primitive types or one of the provided
                MetadataValue types
            logger_defs (Optional[Mapping[str, LoggerDefinition]]):
                A dictionary of string logger identifiers to their implementations.
            executor_def (Optional[ExecutorDefinition]):
                How this Job will be executed. Defaults to :py:class:`multi_or_in_process_executor`,
                which can be switched between multi-process and in-process modes of execution. The
                default mode of execution is multi-process.
            op_retry_policy (Optional[RetryPolicy]): The default retry policy for all ops in this job.
                Only used if retry policy is not defined on the op definition or op invocation.
            partitions_def (Optional[PartitionsDefinition]): Defines a discrete set of partition
                keys that can parameterize the job. If this argument is supplied, the config
                argument can't also be supplied.
            asset_layer (Optional[AssetLayer]): Top level information about the assets this job
                will produce. Generally should not be set manually.
            input_values (Optional[Mapping[str, Any]]):
                A dictionary that maps python objects to the top-level inputs of a job.

        Returns:
            JobDefinition
        """
        from dagster._core.definitions.job_definition import JobDefinition
        from dagster._core.execution.build_resources import wrap_resources_for_execution

        wrapped_resource_defs = wrap_resources_for_execution(resource_defs)
        return JobDefinition.dagster_internal_init(
            name=name,
            description=description or self.description,
            graph_def=self,
            resource_defs=wrapped_resource_defs,
            logger_defs=logger_defs,
            executor_def=executor_def,
            config=config,
            partitions_def=partitions_def,
            tags=tags,
            run_tags=run_tags,
            metadata=metadata,
            hook_defs=hooks,
            op_retry_policy=op_retry_policy,
            asset_layer=asset_layer,
            input_values=input_values,
            _subset_selection_data=_asset_selection_data,
            _was_explicitly_provided_resources=None,  # None means this is determined by whether resource_defs contains any explicitly provided resources
        ).get_subset(op_selection=op_selection)

    def coerce_to_job(self) -> "JobDefinition":
        # attempt to coerce a Graph in to a Job, raising a useful error if it doesn't work
        try:
            return self.to_job()
        except DagsterInvalidDefinitionError as err:
            raise DagsterInvalidDefinitionError(
                f"Failed attempting to coerce Graph {self.name} in to a Job. "
                "Use to_job instead, passing the required information."
            ) from err

    @public
    def execute_in_process(
        self,
        run_config: Any = None,
        instance: Optional["DagsterInstance"] = None,
        resources: Optional[Mapping[str, object]] = None,
        raise_on_error: bool = True,
        op_selection: Optional[Sequence[str]] = None,
        run_id: Optional[str] = None,
        input_values: Optional[Mapping[str, object]] = None,
    ) -> "ExecuteInProcessResult":
        """Execute this graph in-process, collecting results in-memory.

        Args:
            run_config (Optional[Mapping[str, Any]]):
                Run config to provide to execution. The configuration for the underlying graph
                should exist under the "ops" key.
            instance (Optional[DagsterInstance]):
                The instance to execute against, an ephemeral one will be used if none provided.
            resources (Optional[Mapping[str, Any]]):
                The resources needed if any are required. Can provide resource instances directly,
                or resource definitions.
            raise_on_error (Optional[bool]): Whether or not to raise exceptions when they occur.
                Defaults to ``True``.
            op_selection (Optional[List[str]]): A list of op selection queries (including single op
                names) to execute. For example:
                * ``['some_op']``: selects ``some_op`` itself.
                * ``['*some_op']``: select ``some_op`` and all its ancestors (upstream dependencies).
                * ``['*some_op+++']``: select ``some_op``, all its ancestors, and its descendants
                (downstream dependencies) within 3 levels down.
                * ``['*some_op', 'other_op_a', 'other_op_b+']``: select ``some_op`` and all its
                ancestors, ``other_op_a`` itself, and ``other_op_b`` and its direct child ops.
            input_values (Optional[Mapping[str, Any]]):
                A dictionary that maps python objects to the top-level inputs of the graph.

        Returns:
            :py:class:`~dagster.ExecuteInProcessResult`
        """
        from dagster._core.definitions.executor_definition import execute_in_process_executor
        from dagster._core.definitions.job_definition import JobDefinition
        from dagster._core.execution.build_resources import wrap_resources_for_execution
        from dagster._core.instance import DagsterInstance

        instance = check.opt_inst_param(instance, "instance", DagsterInstance)
        resources = check.opt_mapping_param(resources, "resources", key_type=str)
        input_values = check.opt_mapping_param(input_values, "input_values")

        resource_defs = wrap_resources_for_execution(resources)

        ephemeral_job = JobDefinition(
            name=self._name,
            graph_def=self,
            executor_def=execute_in_process_executor,
            resource_defs=resource_defs,
            input_values=input_values,
        ).get_subset(op_selection=op_selection)

        run_config = run_config if run_config is not None else {}
        op_selection = check.opt_sequence_param(op_selection, "op_selection", str)

        return ephemeral_job.execute_in_process(
            run_config=run_config,
            instance=instance,
            raise_on_error=raise_on_error,
            run_id=run_id,
        )

    @property
    def parent_graph_def(self) -> Optional["GraphDefinition"]:
        return None

    @property
    def is_subselected(self) -> bool:
        return False

    def get_resource_requirements(
        self,
        asset_layer: Optional["AssetLayer"],
    ) -> Iterator[ResourceRequirement]:
        for node in self.node_dict.values():
            yield from node.get_resource_requirements(outer_container=self, asset_layer=asset_layer)

        for dagster_type in self.all_dagster_types():
            yield from dagster_type.get_resource_requirements()

    @public
    @property
    def name(self) -> str:
        """The name of the graph."""
        return super().name

    @public
    @property
    def tags(self) -> Mapping[str, str]:
        """The tags associated with the graph."""
        return super().tags

    @property
    def pools(self) -> Set[str]:
        pools = set()
        for node_def in self.node_defs:
            pools.update(node_def.pools)
        return pools

    @public
    def alias(self, name: str) -> "PendingNodeInvocation":
        """Aliases the graph with a new name.

        Can only be used in the context of a :py:func:`@graph <graph>`, :py:func:`@job <job>`, or :py:func:`@asset_graph <asset_graph>` decorated function.

        **Examples:**
            .. code-block:: python

                @job
                def do_it_all():
                    my_graph.alias("my_graph_alias")
        """
        return super().alias(name)

    @public
    def tag(self, tags: Optional[Mapping[str, str]]) -> "PendingNodeInvocation":
        """Attaches the provided tags to the graph immutably.

        Can only be used in the context of a :py:func:`@graph <graph>`, :py:func:`@job <job>`, or :py:func:`@asset_graph <asset_graph>` decorated function.

        **Examples:**
            .. code-block:: python

                @job
                def do_it_all():
                    my_graph.tag({"my_tag": "my_value"})
        """
        return super().tag(tags)

    @public
    def with_hooks(self, hook_defs: AbstractSet[HookDefinition]) -> "PendingNodeInvocation":
        """Attaches the provided hooks to the graph immutably.

        Can only be used in the context of a :py:func:`@graph <graph>`, :py:func:`@job <job>`, or :py:func:`@asset_graph <asset_graph>` decorated function.

        **Examples:**
            .. code-block:: python

                @job
                def do_it_all():
                    my_graph.with_hooks({my_hook})
        """
        return super().with_hooks(hook_defs)

    @public
    def with_retry_policy(self, retry_policy: RetryPolicy) -> "PendingNodeInvocation":
        """Attaches the provided retry policy to the graph immutably.

        Can only be used in the context of a :py:func:`@graph <graph>`, :py:func:`@job <job>`, or :py:func:`@asset_graph <asset_graph>` decorated function.

        **Examples:**
            .. code-block:: python

                @job
                def do_it_all():
                    my_graph.with_retry_policy(RetryPolicy(max_retries=5))
        """
        return super().with_retry_policy(retry_policy)

    def resolve_input_to_destinations(
        self, input_handle: NodeInputHandle
    ) -> Sequence[NodeInputHandle]:
        all_destinations: list[NodeInputHandle] = []
        for mapping in self.input_mappings:
            if mapping.graph_input_name != input_handle.input_name:
                continue
            # recurse into graph structure
            all_destinations += self.node_named(
                mapping.maps_to.node_name
            ).definition.resolve_input_to_destinations(
                NodeInputHandle(
                    node_handle=NodeHandle(
                        mapping.maps_to.node_name, parent=input_handle.node_handle
                    ),
                    input_name=mapping.maps_to.input_name,
                ),
            )

        return all_destinations

    def resolve_output_to_destinations(
        self, output_name: str, handle: Optional[NodeHandle]
    ) -> Sequence[NodeInputHandle]:
        all_destinations: list[NodeInputHandle] = []
        for mapping in self.output_mappings:
            if mapping.graph_output_name != output_name:
                continue
            output_pointer = mapping.maps_from
            output_node = self.node_named(output_pointer.node_name)

            all_destinations.extend(
                output_node.definition.resolve_output_to_destinations(
                    output_pointer.output_name,
                    NodeHandle(output_pointer.node_name, parent=handle),
                )
            )

            output_def = output_node.definition.output_def_named(output_pointer.output_name)
            downstream_input_handles = (
                self.dependency_structure.output_to_downstream_inputs_for_node(
                    output_pointer.node_name
                ).get(NodeOutput(output_node, output_def), [])
            )
            for input_handle in downstream_input_handles:
                all_destinations.append(
                    NodeInputHandle(
                        node_handle=NodeHandle(input_handle.node_name, parent=handle),
                        input_name=input_handle.input_name,
                    )
                )

        return all_destinations

    def get_op_handles(self, parent: NodeHandle) -> AbstractSet[NodeHandle]:
        return {
            op_handle
            for node in self.nodes
            for op_handle in node.definition.get_op_handles(NodeHandle(node.name, parent=parent))
        }

    def get_op_output_handles(self, parent: Optional[NodeHandle]) -> AbstractSet[NodeOutputHandle]:
        return {
            op_output_handle
            for node in self.nodes
            for op_output_handle in node.definition.get_op_output_handles(
                NodeHandle(node.name, parent=parent)
            )
        }

    def get_op_input_output_handle_pairs(
        self, outer_handle: Optional[NodeHandle]
    ) -> AbstractSet[tuple[NodeOutputHandle, NodeInputHandle]]:
        """Get all pairs of op output handles and their downstream op input handles within the graph."""
        result: set[tuple[NodeOutputHandle, NodeInputHandle]] = set()

        for node in self.nodes:
            node_handle = NodeHandle(node.name, parent=outer_handle)
            if isinstance(node.definition, GraphDefinition):
                result.update(node.definition.get_op_input_output_handle_pairs(node_handle))

            for (
                node_input,
                upstream_outputs,
            ) in self.dependency_structure.input_to_upstream_outputs_for_node(node.name).items():
                op_input_handles = node_input.node.definition.resolve_input_to_destinations(
                    NodeInputHandle(node_handle=node_handle, input_name=node_input.input_def.name)
                )
                for op_input_handle in op_input_handles:
                    for upstream_node_output in upstream_outputs:
                        origin_output_def, origin_node_handle = (
                            upstream_node_output.node.definition.resolve_output_to_origin(
                                upstream_node_output.output_def.name,
                                NodeHandle(upstream_node_output.node.name, parent=outer_handle),
                            )
                        )
                        origin_output_handle = NodeOutputHandle(
                            node_handle=origin_node_handle, output_name=origin_output_def.name
                        )

                        result.add((origin_output_handle, op_input_handle))

        return result


class SubselectedGraphDefinition(GraphDefinition):
    """Defines a subselected graph.

    Args:
        parent_graph_def (GraphDefinition): The parent graph that this current graph is subselected
            from. This is used for tracking where the subselected graph originally comes from.
            Note that we allow subselecting a subselected graph, and this field refers to the direct
            parent graph of the current subselection, rather than the original root graph.
        node_defs (Optional[Sequence[NodeDefinition]]): A list of all top level nodes in the graph. A
            node can be an op or a graph that contains other nodes.
        dependencies (Optional[Mapping[Union[str, NodeInvocation], Mapping[str, IDependencyDefinition]]]):
            A structure that declares the dependencies of each op's inputs on the outputs of other
            ops in the subselected graph. Keys of the top level dict are either the string names of
            ops in the graph or, in the case of aliased ops, :py:class:`NodeInvocations <NodeInvocation>`.
            Values of the top level dict are themselves dicts, which map input names belonging to
            the op or aliased op to :py:class:`DependencyDefinitions <DependencyDefinition>`.
        input_mappings (Optional[Sequence[InputMapping]]): Define the inputs to the nested graph, and
            how they map to the inputs of its constituent ops.
        output_mappings (Optional[Sequence[OutputMapping]]): Define the outputs of the nested graph, and
            how they map from the outputs of its constituent ops.
    """

    def __init__(
        self,
        parent_graph_def: GraphDefinition,
        node_defs: Optional[Sequence[NodeDefinition]],
        dependencies: Optional[
            Union[
                DependencyMapping[str],
                DependencyMapping[NodeInvocation],
            ]
        ],
        input_mappings: Optional[Sequence[InputMapping]],
        output_mappings: Optional[Sequence[OutputMapping]],
    ):
        self._parent_graph_def = check.inst_param(
            parent_graph_def, "parent_graph_def", GraphDefinition
        )
        super().__init__(
            name=parent_graph_def.name,  # should we create special name for subselected graphs
            node_defs=node_defs,
            dependencies=dependencies,
            input_mappings=input_mappings,
            output_mappings=output_mappings,
            config=parent_graph_def.config_mapping,
            tags=parent_graph_def.tags,
        )

    @property
    def parent_graph_def(self) -> GraphDefinition:
        return self._parent_graph_def

    def get_top_level_omitted_nodes(self) -> Sequence[Node]:
        return [node for node in self.parent_graph_def.nodes if not self.has_node_named(node.name)]

    @property
    def is_subselected(self) -> bool:
        return True


def _validate_in_mappings(
    input_mappings: Sequence[InputMapping],
    nodes_by_name: Mapping[str, Node],
    dependency_structure: DependencyStructure,
    name: str,
    class_name: str,
) -> Sequence[InputDefinition]:
    from dagster._core.definitions.composition import MappedInputPlaceholder

    input_defs_by_name: dict[str, InputDefinition] = OrderedDict()
    mapping_keys: Set[str] = set()

    target_input_types_by_graph_input_name: dict[str, set[DagsterType]] = defaultdict(set)

    for mapping in input_mappings:
        # handle incorrect objects passed in as mappings
        if not isinstance(mapping, InputMapping):
            if isinstance(mapping, InputDefinition):
                raise DagsterInvalidDefinitionError(
                    f"In {class_name} '{name}' you passed an InputDefinition "
                    f"named '{mapping.name}' directly in to input_mappings. Return "
                    "an InputMapping by calling mapping_to on the InputDefinition."
                )
            else:
                raise DagsterInvalidDefinitionError(
                    f"In {class_name} '{name}' received unexpected type '{type(mapping)}' in"
                    " input_mappings. Provide an InputMapping using InputMapping(...)"
                )

        input_defs_by_name[mapping.graph_input_name] = mapping.get_definition()

        target_node = nodes_by_name.get(mapping.maps_to.node_name)
        if target_node is None:
            raise DagsterInvalidDefinitionError(
                f"In {class_name} '{name}' input mapping references node "
                f"'{mapping.maps_to.node_name}' which it does not contain."
            )
        if not target_node.has_input(mapping.maps_to.input_name):
            raise DagsterInvalidDefinitionError(
                f"In {class_name} '{name}' input mapping to node '{mapping.maps_to.node_name}' "
                f"which contains no input named '{mapping.maps_to.input_name}'"
            )

        target_input_def = target_node.input_def_named(mapping.maps_to.input_name)
        node_input = NodeInput(target_node, target_input_def)

        if mapping.maps_to_fan_in:
            maps_to = cast(FanInInputPointer, mapping.maps_to)
            if not dependency_structure.has_fan_in_deps(node_input):
                raise DagsterInvalidDefinitionError(
                    f"In {class_name} '{name}' input mapping target"
                    f' "{maps_to.node_name}.{maps_to.input_name}" (index'
                    f" {maps_to.fan_in_index} of fan-in) is not a MultiDependencyDefinition."
                )
            inner_deps = dependency_structure.get_fan_in_deps(node_input)
            if (maps_to.fan_in_index >= len(inner_deps)) or (
                inner_deps[maps_to.fan_in_index] is not MappedInputPlaceholder
            ):
                raise DagsterInvalidDefinitionError(
                    f"In {class_name} '{name}' input mapping target "
                    f'"{maps_to.node_name}.{maps_to.input_name}" index {maps_to.fan_in_index} in '
                    "the MultiDependencyDefinition is not a MappedInputPlaceholder"
                )
            mapping_keys.add(f"{maps_to.node_name}.{maps_to.input_name}.{maps_to.fan_in_index}")
            target_input_types_by_graph_input_name[mapping.graph_input_name].add(
                target_input_def.dagster_type.get_inner_type_for_fan_in()
            )
        else:
            if dependency_structure.has_deps(node_input):
                raise DagsterInvalidDefinitionError(
                    f"In {class_name} '{name}' input mapping target "
                    f'"{mapping.maps_to.node_name}.{mapping.maps_to.input_name}" '
                    "is already satisfied by output"
                )

            mapping_keys.add(f"{mapping.maps_to.node_name}.{mapping.maps_to.input_name}")
            target_input_types_by_graph_input_name[mapping.graph_input_name].add(
                target_input_def.dagster_type
            )

    for node_input in dependency_structure.inputs():
        if dependency_structure.has_fan_in_deps(node_input):
            for idx, dep in enumerate(dependency_structure.get_fan_in_deps(node_input)):
                if dep is MappedInputPlaceholder:
                    mapping_str = f"{node_input.node_name}.{node_input.input_name}.{idx}"
                    if mapping_str not in mapping_keys:
                        raise DagsterInvalidDefinitionError(
                            f"Unsatisfied MappedInputPlaceholder at index {idx} in"
                            " MultiDependencyDefinition for"
                            f" '{node_input.node_name}.{node_input.input_name}'"
                        )

    # if the dagster type on a graph input is Any and all its target inputs have the
    # same dagster type, then use that dagster type for the graph input
    for graph_input_name, graph_input_def in input_defs_by_name.items():
        if graph_input_def.dagster_type.kind == DagsterTypeKind.ANY:
            target_input_types = target_input_types_by_graph_input_name[graph_input_name]
            if len(target_input_types) == 1:
                input_defs_by_name[graph_input_name] = graph_input_def.with_dagster_type(
                    next(iter(target_input_types))
                )

    return list(input_defs_by_name.values())


def _validate_out_mappings(
    output_mappings: Sequence[OutputMapping],
    node_dict: Mapping[str, Node],
    name: str,
    class_name: str,
) -> tuple[Sequence[OutputMapping], Sequence[OutputDefinition]]:
    output_defs: list[OutputDefinition] = []
    for mapping in output_mappings:
        if isinstance(mapping, OutputMapping):
            target_node = node_dict.get(mapping.maps_from.node_name)
            if target_node is None:
                raise DagsterInvalidDefinitionError(
                    f"In {class_name} '{name}', output mapping references node "
                    f"'{mapping.maps_from.node_name}' which it does not contain."
                )
            if not target_node.has_output(mapping.maps_from.output_name):
                raise DagsterInvalidDefinitionError(
                    f"In {class_name} {name}, output mapping from {target_node.describe_node()} "
                    f"references output '{mapping.maps_from.output_name}', which the node does not "
                    "contain."
                )

            target_output = target_node.output_def_named(mapping.maps_from.output_name)
            output_def = mapping.get_definition(is_dynamic=target_output.is_dynamic)
            output_defs.append(output_def)

            if (
                mapping.dagster_type
                and mapping.dagster_type.kind != DagsterTypeKind.ANY
                and (target_output.dagster_type != mapping.dagster_type)
                and class_name != "GraphDefinition"
            ):
                raise DagsterInvalidDefinitionError(
                    f"In {class_name} '{name}' output '{mapping.graph_output_name}' of type"
                    f" {mapping.dagster_type.display_name} maps from"
                    f" {mapping.maps_from.node_name}.{mapping.maps_from.output_name} of different"
                    f" type {target_output.dagster_type.display_name}. OutputMapping source and"
                    " destination must have the same type."
                )

        elif isinstance(mapping, OutputDefinition):
            raise DagsterInvalidDefinitionError(
                f"You passed an OutputDefinition named '{mapping.name}' directly "
                "in to output_mappings. Return an OutputMapping by calling "
                "mapping_from on the OutputDefinition."
            )
        else:
            raise DagsterInvalidDefinitionError(
                f"Received unexpected type '{type(mapping)}' in output_mappings. "
                "Provide an OutputMapping using OutputDefinition(...).mapping_from(...)"
            )
    return output_mappings, output_defs



from collections.abc import Iterable, Mapping
from enum import Enum
from typing import TYPE_CHECKING, Any, NamedTuple, Optional, Union

import dagster._check as check
from dagster._annotations import PublicAttr
from dagster._core.definitions.asset_key import AssetCheckKey, AssetKey, CoercibleToAssetKey
from dagster._serdes.serdes import whitelist_for_serdes

if TYPE_CHECKING:
    from dagster._core.definitions.asset_dep import AssetDep, CoercibleToAssetDep
    from dagster._core.definitions.assets import AssetsDefinition
    from dagster._core.definitions.declarative_automation.automation_condition import (
        AutomationCondition,
    )
    from dagster._core.definitions.source_asset import SourceAsset


@whitelist_for_serdes
class AssetCheckSeverity(Enum):
    """Severity level for an AssetCheckResult.

    - WARN: a potential issue with the asset
    - ERROR: a definite issue with the asset

    Severity does not impact execution of the asset or downstream assets.
    """

    WARN = "WARN"
    ERROR = "ERROR"


class AssetCheckSpec(
    NamedTuple(
        "_AssetCheckSpec",
        [
            ("name", PublicAttr[str]),
            ("asset_key", PublicAttr[AssetKey]),
            ("description", PublicAttr[Optional[str]]),
            ("additional_deps", PublicAttr[Iterable["AssetDep"]]),
            (
                "blocking",  # intentionally not public, see https://github.com/dagster-io/dagster/issues/20659
                bool,
            ),
            ("metadata", PublicAttr[Optional[Mapping[str, Any]]]),
            ("automation_condition", Optional["AutomationCondition[AssetCheckKey]"]),
        ],
    )
):
    """Defines information about an asset check, except how to execute it.

    AssetCheckSpec is often used as an argument to decorators that decorator a function that can
    execute multiple checks - e.g. `@asset`, and `@multi_asset`. It defines one of the checks that
    will be executed inside that function.

    Args:
        name (str): Name of the check.
        asset (Union[AssetKey, Sequence[str], str, AssetsDefinition, SourceAsset]): The asset that
            the check applies to.
        description (Optional[str]): Description for the check.
        additional_deps (Optional[Iterable[AssetDep]]): Additional dependencies for the check. The
            check relies on these assets in some way, but the result of the check only applies to
            the asset specified by `asset`. For example, the check may test that `asset` has
            matching data with an asset in `additional_deps`. This field holds both `additional_deps`
            and `additional_ins` passed to @asset_check.
        metadata (Optional[Mapping[str, Any]]):  A dict of static metadata for this asset check.
    """

    def __new__(
        cls,
        name: str,
        *,
        asset: Union[CoercibleToAssetKey, "AssetsDefinition", "SourceAsset"],
        description: Optional[str] = None,
        additional_deps: Optional[Iterable["CoercibleToAssetDep"]] = None,
        blocking: bool = False,
        metadata: Optional[Mapping[str, Any]] = None,
        automation_condition: Optional["AutomationCondition[AssetCheckKey]"] = None,
    ):
        from dagster._core.definitions.asset_dep import coerce_to_deps_and_check_duplicates
        from dagster._core.definitions.declarative_automation.automation_condition import (
            AutomationCondition,
        )

        asset_key = AssetKey.from_coercible_or_definition(asset)

        additional_asset_deps = coerce_to_deps_and_check_duplicates(
            additional_deps, AssetCheckKey(asset_key, name)
        )

        for dep in additional_asset_deps:
            if dep.asset_key == asset_key:
                raise ValueError(
                    f"Asset check {name} for asset {asset_key.to_string()} cannot have an additional "
                    f"dependency on asset {asset_key.to_string()}."
                )

        return super().__new__(
            cls,
            name=check.str_param(name, "name"),
            asset_key=asset_key,
            description=check.opt_str_param(description, "description"),
            additional_deps=additional_asset_deps,
            blocking=check.bool_param(blocking, "blocking"),
            metadata=check.opt_mapping_param(metadata, "metadata", key_type=str),
            automation_condition=check.opt_inst_param(
                automation_condition, "automation_condition", AutomationCondition
            ),
        )

    def get_python_identifier(self) -> str:
        """Returns a string uniquely identifying the asset check, that uses only the characters
        allowed in a Python identifier.
        """
        return f"{self.asset_key.to_python_identifier()}_{self.name}".replace(".", "_")

    @property
    def key(self) -> AssetCheckKey:
        return AssetCheckKey(self.asset_key, self.name)

    def replace_key(self, key: AssetCheckKey) -> "AssetCheckSpec":
        return self._replace(asset_key=key.asset_key, name=key.name)



"""add run job index

Revision ID: 16689497301f
Revises: 6df03f4b1efb
Create Date: 2023-02-01 11:27:14.146322

"""

from dagster._core.storage.migration.utils import add_run_job_index, drop_run_job_index

# revision identifiers, used by Alembic.
revision = "16689497301f"
down_revision = "6df03f4b1efb"
branch_labels = None
depends_on = None


def upgrade():
    add_run_job_index()


def downgrade():
    drop_run_job_index()



import dataclasses
import datetime
import functools
import logging
import os
from collections import defaultdict
from collections.abc import Mapping, Sequence
from dataclasses import dataclass, replace
from typing import TYPE_CHECKING, AbstractSet, Any, Callable, Optional, TypeVar  # noqa: UP035

import dagster._check as check
from dagster._core.asset_graph_view.entity_subset import EntitySubset
from dagster._core.asset_graph_view.serializable_entity_subset import SerializableEntitySubset
from dagster._core.definitions.asset_key import EntityKey
from dagster._core.definitions.declarative_automation.legacy.valid_asset_subset import (
    ValidAssetSubset,
)
from dagster._core.definitions.declarative_automation.serialized_objects import (
    AssetSubsetWithMetadata,
    AutomationConditionCursor,
    AutomationConditionNodeCursor,
    HistoricalAllPartitionsSubsetSentinel,
)
from dagster._core.definitions.events import AssetKey, AssetKeyPartitionKey
from dagster._core.definitions.metadata import MetadataValue
from dagster._core.definitions.partition import PartitionsDefinition
from dagster._time import get_current_timestamp

if TYPE_CHECKING:
    from dagster._core.definitions.base_asset_graph import BaseAssetGraph
    from dagster._core.definitions.data_time import CachingDataTimeResolver
    from dagster._core.definitions.declarative_automation.automation_condition import (
        AutomationCondition,
    )
    from dagster._core.definitions.declarative_automation.automation_condition_evaluator import (
        AutomationConditionEvaluator,
    )
    from dagster._utils.caching_instance_queryer import CachingInstanceQueryer

T = TypeVar("T")


def root_property(fn: Callable[[Any], T]) -> Callable[[Any], T]:
    """Ensures that a given property is always accessed via the root context, ensuring that any
    cached properties are accessed correctly.
    """

    def wrapped(self: Any) -> Any:
        return fn(self.root_context)

    return wrapped


@dataclass(frozen=True)
class LegacyRuleEvaluationContext:
    """Context object containing methods and properties used for evaluating the entire state of an
    asset's automation rules.
    """

    asset_key: AssetKey
    condition: "AutomationCondition"
    cursor: Optional[AutomationConditionCursor]
    node_cursor: Optional[AutomationConditionNodeCursor]
    candidate_subset: ValidAssetSubset

    instance_queryer: "CachingInstanceQueryer"
    data_time_resolver: "CachingDataTimeResolver"

    request_subsets_by_key: Mapping[EntityKey, EntitySubset]
    expected_data_time_mapping: Mapping[AssetKey, Optional[datetime.datetime]]

    start_timestamp: float
    respect_materialization_data_versions: bool
    auto_materialize_run_tags: Mapping[str, str]
    logger: logging.Logger
    root_ref: Optional["LegacyRuleEvaluationContext"] = None

    @staticmethod
    def create(asset_key: AssetKey, evaluator: "AutomationConditionEvaluator"):
        instance_queryer = evaluator.asset_graph_view.get_inner_queryer_for_back_compat()

        cursor = evaluator.cursor.get_previous_condition_cursor(asset_key)
        condition = check.not_none(evaluator.asset_graph.get(asset_key).automation_condition)
        partitions_def = evaluator.asset_graph.get(asset_key).partitions_def

        return LegacyRuleEvaluationContext(
            asset_key=asset_key,
            condition=condition,
            cursor=cursor,
            node_cursor=cursor.node_cursors_by_unique_id.get(
                condition.get_node_unique_id(parent_unique_id=None, index=0)
            )
            if cursor
            else None,
            candidate_subset=ValidAssetSubset.all(
                asset_key,
                partitions_def,
                instance_queryer,
                instance_queryer.evaluation_time,
            ),
            data_time_resolver=evaluator.legacy_data_time_resolver,
            instance_queryer=instance_queryer,
            request_subsets_by_key=evaluator.request_subsets_by_key,
            expected_data_time_mapping=evaluator.legacy_expected_data_time_by_key,
            start_timestamp=get_current_timestamp(),
            respect_materialization_data_versions=evaluator.legacy_respect_materialization_data_versions,
            auto_materialize_run_tags=evaluator.legacy_auto_materialize_run_tags,
            logger=evaluator.logger,
        )

    def for_child(
        self,
        child_condition: "AutomationCondition",
        child_unique_id: str,
        candidate_subset: EntitySubset,
    ) -> "LegacyRuleEvaluationContext":
        return dataclasses.replace(
            self,
            condition=child_condition,
            node_cursor=self.cursor.node_cursors_by_unique_id.get(child_unique_id)
            if self.cursor
            else None,
            candidate_subset=ValidAssetSubset(
                key=candidate_subset.key, value=candidate_subset.get_internal_value()
            ),
            root_ref=self.root_context,
            start_timestamp=get_current_timestamp(),
        )

    @property
    def root_context(self) -> "LegacyRuleEvaluationContext":
        """A reference to the context of the root condition for this evaluation."""
        return self.root_ref or self

    @property
    def asset_graph(self) -> "BaseAssetGraph":
        return self.instance_queryer.asset_graph

    @property
    def partitions_def(self) -> Optional[PartitionsDefinition]:
        return self.asset_graph.get(self.asset_key).partitions_def

    @property
    def evaluation_time(self) -> datetime.datetime:
        """Returns the time at which this rule is being evaluated."""
        return self.instance_queryer.evaluation_time

    @property
    def previous_max_storage_id(self) -> Optional[int]:
        return self.cursor.temporal_context.last_event_id if self.cursor else None

    @property
    def previous_evaluation_timestamp(self) -> Optional[float]:
        return self.cursor.effective_timestamp if self.cursor else None

    @property
    def previous_true_subset(self) -> SerializableEntitySubset:
        if self.node_cursor is None:
            return self.empty_subset()
        return self.node_cursor.true_subset

    @property
    def previous_candidate_subset(self) -> SerializableEntitySubset:
        if self.node_cursor is None:
            return self.empty_subset()
        candidate_subset = self.node_cursor.candidate_subset
        if isinstance(candidate_subset, HistoricalAllPartitionsSubsetSentinel):
            return ValidAssetSubset.all(
                self.asset_key, self.partitions_def, self.instance_queryer, self.evaluation_time
            )
        else:
            return candidate_subset

    @property
    def previous_subsets_with_metadata(self) -> Sequence[AssetSubsetWithMetadata]:
        if self.node_cursor is None:
            return []
        return self.node_cursor.subsets_with_metadata

    @functools.cached_property
    @root_property
    def parent_will_update_subset(self) -> ValidAssetSubset:
        """Returns the set of asset partitions whose parents will be updated on this tick, and which
        can be materialized in the same run as this asset.
        """
        subset = self.empty_subset()
        for parent_key in self.asset_graph.get(self.asset_key).parent_keys:
            if not self.materializable_in_same_run(self.asset_key, parent_key):
                continue
            parent_subset = self.request_subsets_by_key.get(parent_key)
            if not parent_subset:
                continue
            parent_subset = ValidAssetSubset.coerce_from_subset(
                parent_subset.convert_to_serializable_subset(), self.partitions_def
            )
            subset |= replace(parent_subset, key=self.asset_key)
        return subset

    @functools.cached_property
    @root_property
    def materialized_since_previous_tick_subset(self) -> ValidAssetSubset:
        """Returns the set of asset partitions that were materialized since the previous tick."""
        return ValidAssetSubset.from_asset_partitions_set(
            self.asset_key,
            self.partitions_def,
            self.instance_queryer.get_asset_partitions_updated_after_cursor(
                self.asset_key,
                asset_partitions=None,
                after_cursor=self.previous_max_storage_id,
                respect_materialization_data_versions=False,
            ),
        )

    @property
    @root_property
    def _previous_tick_discarded_subset(self) -> Optional[SerializableEntitySubset[AssetKey]]:
        """Fetches the unique id corresponding to the DiscardOnMaxMaterializationsExceededRule, if
        that rule is part of the broader condition.
        """
        from dagster._core.definitions.auto_materialize_rule_impls import (
            DiscardOnMaxMaterializationsExceededRule,
        )
        from dagster._core.definitions.declarative_automation.legacy.rule_condition import (
            RuleCondition,
        )
        from dagster._core.definitions.declarative_automation.operators import (
            NotAutomationCondition,
        )

        # if you have a discard condition, it'll be part of a structure of the form
        # Or(MaterializeCond, Not(SkipCond), Not(DiscardCond))
        if len(self.condition.children) != 3:
            return None
        unique_id = self.condition.get_node_unique_id(parent_unique_id=None, index=None)

        # get Not(DiscardCond)
        not_discard_condition = self.condition.children[2]
        unique_id = not_discard_condition.get_node_unique_id(parent_unique_id=unique_id, index=2)
        if not isinstance(not_discard_condition, NotAutomationCondition):
            return None

        # get DiscardCond
        discard_condition = not_discard_condition.children[0]
        unique_id = discard_condition.get_node_unique_id(parent_unique_id=unique_id, index=0)
        if not isinstance(discard_condition, RuleCondition) or not isinstance(
            discard_condition.rule, DiscardOnMaxMaterializationsExceededRule
        ):
            return None

        # grab the stored cursor value for the discard condition
        discard_cursor = (
            self.cursor.node_cursors_by_unique_id.get(unique_id) if self.cursor else None
        )
        return discard_cursor.true_subset if discard_cursor else None

    @property
    @root_property
    def previous_tick_requested_subset(self) -> SerializableEntitySubset:
        """The set of asset partitions that were requested (or discarded) on the previous tick."""
        if self.cursor is None:
            return self.empty_subset()

        discarded_subset = self._previous_tick_discarded_subset
        requested_subset = self.cursor.previous_requested_subset
        return (
            ValidAssetSubset.coerce_from_subset(requested_subset, self.partitions_def)
            | discarded_subset
            if discarded_subset
            else requested_subset
        )

    @property
    def materialized_requested_or_discarded_since_previous_tick_subset(self) -> ValidAssetSubset:
        """Returns the set of asset partitions that were materialized since the previous tick."""
        return self.materialized_since_previous_tick_subset | self.previous_tick_requested_subset

    @functools.cached_property
    @root_property
    def _parent_has_updated_subset_and_new_latest_storage_id(
        self,
    ) -> tuple[ValidAssetSubset, Optional[int]]:
        """Returns the set of asset partitions whose parents have updated since the last time this
        condition was evaluated.
        """
        max_child_partitions_str = os.getenv("DAGSTER_MAX_AMP_CHILD_PARTITIONS", None)
        (
            asset_partitions,
            cursor,
        ) = self.root_context.instance_queryer.asset_partitions_with_newly_updated_parents_and_new_cursor(
            latest_storage_id=self.previous_max_storage_id,
            child_asset_key=self.root_context.asset_key,
            map_old_time_partitions=False,
            max_child_partitions=int(max_child_partitions_str)
            if max_child_partitions_str
            else None,
        )
        return ValidAssetSubset.from_asset_partitions_set(
            self.asset_key, self.partitions_def, asset_partitions
        ), cursor

    @property
    @root_property
    def parent_has_updated_subset(self) -> ValidAssetSubset:
        subset, _ = self._parent_has_updated_subset_and_new_latest_storage_id
        return subset

    @property
    @root_property
    def new_max_storage_id(self) -> Optional[int]:
        _, storage_id = self._parent_has_updated_subset_and_new_latest_storage_id
        return storage_id

    @property
    def parent_has_or_will_update_subset(self) -> ValidAssetSubset:
        """Returns the set of asset partitions whose parents have updated since the last time this
        condition was evaluated, or will update on this tick.
        """
        return self.parent_has_updated_subset | self.root_context.parent_will_update_subset

    @property
    def candidate_parent_has_or_will_update_subset(self) -> ValidAssetSubset:
        """Returns the set of candidates for this tick which have parents that have updated since
        the previous tick, or will update on this tick.
        """
        return self.candidate_subset & self.parent_has_or_will_update_subset

    @property
    def candidates_not_evaluated_on_previous_tick_subset(self) -> ValidAssetSubset:
        """Returns the set of candidates for this tick which were not candidates on the previous
        tick.
        """
        from dagster._core.definitions.declarative_automation.serialized_objects import (
            HistoricalAllPartitionsSubsetSentinel,
        )

        if not self.node_cursor:
            return self.candidate_subset
        # when the candidate_subset is HistoricalAllPartitionsSubsetSentinel, this indicates that the
        # entire asset was evaluated for this condition on the previous tick, and so no candidates
        # were *not* evaluated on the previous tick
        elif isinstance(self.node_cursor.candidate_subset, HistoricalAllPartitionsSubsetSentinel):
            return self.empty_subset()
        return self.candidate_subset - self.node_cursor.candidate_subset

    def materializable_in_same_run(self, child_key: AssetKey, parent_key: AssetKey) -> bool:
        """Returns whether a child asset can be materialized in the same run as a parent asset."""
        from dagster._core.definitions.asset_graph import executable_in_same_run

        return executable_in_same_run(self.asset_graph, child_key, parent_key)

    def get_parents_that_will_not_be_materialized_on_current_tick(
        self, *, asset_partition: AssetKeyPartitionKey
    ) -> AbstractSet[AssetKeyPartitionKey]:
        """Returns the set of parent asset partitions that will not be updated in the same run of
        this asset partition if a run is launched for this asset partition on this tick.
        """
        return {
            parent
            for parent in self.asset_graph.get_parents_partitions(
                dynamic_partitions_store=self.instance_queryer,
                current_time=self.instance_queryer.evaluation_time,
                asset_key=asset_partition.asset_key,
                partition_key=asset_partition.partition_key,
            ).parent_partitions
            if not self.will_update_asset_partition(parent)
            or not self.materializable_in_same_run(asset_partition.asset_key, parent.asset_key)
        }

    def will_update_asset_partition(self, asset_partition: AssetKeyPartitionKey) -> bool:
        parent_subset = self.request_subsets_by_key.get(asset_partition.asset_key)
        if not parent_subset:
            return False
        return asset_partition in parent_subset.convert_to_serializable_subset()

    def add_evaluation_data_from_previous_tick(
        self,
        asset_partitions_by_frozen_metadata: Mapping[
            frozenset[tuple[str, MetadataValue]], AbstractSet[AssetKeyPartitionKey]
        ],
        ignore_subset: SerializableEntitySubset,
    ) -> tuple[ValidAssetSubset, Sequence[AssetSubsetWithMetadata]]:
        """Combines information calculated on this tick with information from the previous tick,
        returning a tuple of the combined true subset and the combined subsets with metadata.

        Args:
            asset_partitions_by_frozen_metadata: A mapping from metadata to the set of asset
                partitions that the rule applies to.
            ignore_subset: An EntitySubset which represents information that we should *not* carry
                forward from the previous tick.
        """
        from dagster._core.definitions.declarative_automation.serialized_objects import (
            AssetSubsetWithMetadata,
        )

        mapping = defaultdict(lambda: self.empty_subset())
        has_new_metadata_subset = self.empty_subset()
        for frozen_metadata, asset_partitions in asset_partitions_by_frozen_metadata.items():
            mapping[frozen_metadata] = ValidAssetSubset.from_asset_partitions_set(
                self.asset_key, self.partitions_def, asset_partitions
            )
            has_new_metadata_subset |= mapping[frozen_metadata]

        # don't use information from the previous tick if we have explicit metadata for it or
        # we've explicitly said to ignore it
        ignore_subset = has_new_metadata_subset | ignore_subset

        for elt in self.previous_subsets_with_metadata:
            carry_forward_subset = (
                ValidAssetSubset.coerce_from_subset(elt.subset, self.partitions_def) - ignore_subset
            )
            if carry_forward_subset.size > 0:
                mapping[elt.frozen_metadata] |= carry_forward_subset

        # for now, an asset is in the "true" subset if and only if we have some metadata for it
        true_subset = self.empty_subset()
        for subset in mapping.values():
            true_subset |= subset

        return (
            self.candidate_subset & true_subset,
            [
                AssetSubsetWithMetadata(subset=subset, metadata=dict(metadata))
                for metadata, subset in mapping.items()
            ],
        )

    def empty_subset(self) -> ValidAssetSubset:
        return ValidAssetSubset.empty(self.asset_key, self.partitions_def)



from collections.abc import Iterator, Mapping
from contextlib import contextmanager
from typing import Any, Callable, Generic, Optional, TypeVar

from dagster import _check as check
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.definitions.definitions_load_context import (
    DefinitionsLoadContext,
    DefinitionsLoadType,
)
from dagster._core.definitions.metadata.metadata_value import (
    CodeLocationReconstructionMetadataValue,
)
from dagster._core.definitions.repository_definition.repository_definition import (
    RepositoryDefinition,
    RepositoryLoadData,
)
from dagster._core.errors import DagsterInvariantViolationError
from dagster._record import record
from dagster._serdes.serdes import serialize_value

T_Defs = TypeVar("T_Defs", Definitions, RepositoryDefinition)


@record
class LazyDefinitions(Generic[T_Defs]):
    """An object that can be invoked to load a set of definitions. Useful in tests when you want to regenerate the same definitions in multiple contexts."""

    load_fn: Callable[[], T_Defs]

    def __call__(self) -> T_Defs:
        """Load a set of definitions using the load_fn provided at construction time.

        Returns:
            Union[Definitions, RepositoryDefinition]: The loaded definitions.
        """
        result = self.load_fn()
        if not isinstance(result, (Definitions, RepositoryDefinition)):
            raise DagsterInvariantViolationError(
                "DefinitionsLoader must return a Definitions or RepositoryDefinition object"
            )
        return result


def lazy_definitions(fn: Callable[[], T_Defs]) -> LazyDefinitions[T_Defs]:
    """Marks a function as an entry point for loading a set of Dagster definitions. Useful as a test
    utility to define definitions that you wish to load multiple times with different contexts.

    As with plain `Definitions` objects, there can be only one `@lazy_definitions`-decorated function per
    module if that module is being loaded as a Dagster code location.

    Returns:
        LazyDefinitions: A callable that will load a set of definitions when invoked.

    Examples:
        .. code-block:: python

            from dagster import (
                AssetSpec,
                Definitions,
                DefinitionsLoadContext,
                DefinitionsLoadType,
                asset,
                definitions,
                external_assets_from_specs,
            )

            WORKSPACE_ID = "my_workspace"
            FOO_METADATA_KEY_PREFIX = "foo"


            # Simple model of an external service foo
            def fetch_foo_defs_metadata(workspace_id: str):
                if workspace_id == WORKSPACE_ID:
                    return [{"id": "alpha"}, {"id": "beta"}]
                else:
                    raise Exception("Unknown workspace")


            def get_foo_defs(context: DefinitionsLoadContext, workspace_id: str) -> Definitions:
                metadata_key = f"{FOO_METADATA_KEY_PREFIX}/{workspace_id}"
                if (
                    context.load_type == DefinitionsLoadType.RECONSTRUCTION
                    and metadata_key in context.reconstruction_metadata
                ):
                    payload = context.reconstruction_metadata[metadata_key]
                else:
                    payload = fetch_foo_defs_metadata(workspace_id)
                asset_specs = [AssetSpec(item["id"]) for item in payload]
                assets = external_assets_from_specs(asset_specs)
                return Definitions(
                    assets=assets,
                ).with_reconstruction_metadata({metadata_key: payload})


            @lazy_definitions
            def defs():
                @asset
                def regular_asset(): ...

                context = DefinitionsLoadContext.get()

                return Definitions.merge(
                    get_foo_defs(context, WORKSPACE_ID),
                    Definitions(assets=[regular_asset]),
                )

    """
    return LazyDefinitions(load_fn=fn)


@contextmanager
def scoped_definitions_load_context(
    load_type: DefinitionsLoadType = DefinitionsLoadType.INITIALIZATION,
    repository_load_data: Optional[RepositoryLoadData] = None,
) -> Iterator["DefinitionsLoadContext"]:
    context = DefinitionsLoadContext(load_type=load_type, repository_load_data=repository_load_data)
    curr_context = DefinitionsLoadContext.get()
    try:
        DefinitionsLoadContext.set(context)
        yield context
    finally:
        DefinitionsLoadContext.set(curr_context)


@contextmanager
def scoped_reconstruction_serdes_objects(
    state_objects: Optional[Mapping[str, Any]] = None,
) -> Iterator[None]:
    """For test cases that involved state-backed definition objects. This context manager
    allows one to set backing state for a definitions object and test reconstruction
    logic. Creates a DefinitionsLoadContext with DefinitionsLoadType.RECONSTRUCTION,
    serializes the passed in objects and wraps them in
    a CodeLocationReconstructionMetadataValue on your behalf.

    Examples:

    .. code-block:: python

        with scoped_reconstruction_serdes_objects(test_key=ExampleDefState(a_string="bar")):
            loader_cached = ExampleStateBackedDefinitionsLoader("test_key")
            defs = loader_cached.build_defs()
            assert len(defs.get_all_asset_specs()) == 1
            assert defs.get_assets_def("bar")
    """
    with scoped_reconstruction_metadata(
        {k: serialize_value(v) for k, v in state_objects.items()} if state_objects else None
    ):
        yield


@contextmanager
def scoped_reconstruction_metadata(
    reconstruction_metadata: Optional[Mapping[str, Any]] = None,
) -> Iterator[None]:
    """For test cases that involved state-backed definition objects. This context manager
    allows one to set backing reconstruciton metaddata for a definitions object
    and test reconstruction logic. Creates a DefinitionsLoadContext with
    DefinitionsLoadType.RECONSTRUCTION. Wraps each passed value in a
    CodeLocationReconstructionMetadataValue.

    Examples:

    .. code-block:: python

        with scoped_reconstruction_metadata({"test_key": "test_value"}):
            loader_cached = ExampleStateBackedDefinitionsLoader("test_key")
            defs = loader_cached.build_defs()
            assert len(defs.get_all_asset_specs()) == 1
            assert defs.get_assets_def("bar")
    """
    prev_context = DefinitionsLoadContext.get()
    reconstruction_metadata = reconstruction_metadata or {}
    try:
        prev_load_data = prev_context._repository_load_data  # noqa
        DefinitionsLoadContext.set(
            DefinitionsLoadContext(
                load_type=DefinitionsLoadType.RECONSTRUCTION,
                repository_load_data=RepositoryLoadData(
                    cacheable_asset_data=prev_load_data.cacheable_asset_data
                    if prev_load_data
                    else {},
                    reconstruction_metadata={
                        **{
                            k: CodeLocationReconstructionMetadataValue(v)
                            for k, v in reconstruction_metadata.items()
                        },
                        **(prev_load_data.reconstruction_metadata if prev_load_data else {}),
                    },
                ),
            )
        )
        yield
    finally:
        DefinitionsLoadContext.set(prev_context)


def unwrap_reconstruction_metadata(repo_def: RepositoryDefinition) -> Mapping[str, Any]:
    """Takes the metadata from a Definitions object and unwraps the CodeLocationReconstructionMetadataValue
    metadata values into a dictionary.
    """
    return {
        k: metadata_value.value
        for k, metadata_value in check.not_none(
            repo_def.repository_load_data
        ).reconstruction_metadata.items()
        if isinstance(metadata_value, CodeLocationReconstructionMetadataValue)
    }



import inspect
from collections.abc import Iterable, Iterator, Sequence
from importlib import import_module
from types import ModuleType
from typing import Optional, Union, cast, get_args

import dagster._check as check
from dagster._core.definitions.asset_checks import has_only_asset_checks
from dagster._core.definitions.asset_key import (
    CoercibleToAssetKeyPrefix,
    check_opt_coercible_to_asset_key_prefix_param,
)
from dagster._core.definitions.asset_spec import AssetSpec
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.auto_materialize_policy import AutoMaterializePolicy
from dagster._core.definitions.backfill_policy import BackfillPolicy
from dagster._core.definitions.cacheable_assets import CacheableAssetsDefinition
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
)
from dagster._core.definitions.freshness_policy import FreshnessPolicy
from dagster._core.definitions.module_loaders.object_list import ModuleScopedDagsterDefs
from dagster._core.definitions.module_loaders.utils import find_modules_in_package
from dagster._core.definitions.source_asset import SourceAsset
from dagster._core.definitions.utils import resolve_automation_condition


def find_objects_in_module_of_types(
    module: ModuleType,
    types: Union[type, tuple[type, ...]],
) -> Iterator:
    """Yields instances or subclasses of the given type(s)."""
    for attr in dir(module):
        value = getattr(module, attr)
        if isinstance(value, types):
            yield value
        elif isinstance(value, list) and all(isinstance(el, types) for el in value):
            yield from value


def find_subclasses_in_module(
    module: ModuleType,
    types: Union[type, tuple[type, ...]],
) -> Iterator:
    """Yields instances or subclasses of the given type(s)."""
    for attr in dir(module):
        value = getattr(module, attr)
        if isinstance(value, type) and issubclass(value, types):
            yield value


AssetLoaderTypes = Union[AssetsDefinition, SourceAsset, CacheableAssetsDefinition, AssetSpec]


def load_assets_from_modules(
    modules: Iterable[ModuleType],
    group_name: Optional[str] = None,
    key_prefix: Optional[CoercibleToAssetKeyPrefix] = None,
    *,
    freshness_policy: Optional[FreshnessPolicy] = None,
    auto_materialize_policy: Optional[AutoMaterializePolicy] = None,
    automation_condition: Optional[AutomationCondition] = None,
    backfill_policy: Optional[BackfillPolicy] = None,
    source_key_prefix: Optional[CoercibleToAssetKeyPrefix] = None,
    include_specs: bool = False,
) -> Sequence[AssetLoaderTypes]:
    """Constructs a list of assets and source assets from the given modules.

    Args:
        modules (Iterable[ModuleType]): The Python modules to look for assets inside.
        group_name (Optional[str]):
            Group name to apply to the loaded assets. The returned assets will be copies of the
            loaded objects, with the group name added.
        key_prefix (Optional[Union[str, Sequence[str]]]):
            Prefix to prepend to the keys of the loaded assets. The returned assets will be copies
            of the loaded objects, with the prefix prepended.
        freshness_policy (Optional[FreshnessPolicy]): FreshnessPolicy to apply to all the loaded
            assets.
        automation_condition (Optional[AutomationCondition]): AutomationCondition to apply
            to all the loaded assets.
        backfill_policy (Optional[AutoMaterializePolicy]): BackfillPolicy to apply to all the loaded assets.
        source_key_prefix (bool): Prefix to prepend to the keys of loaded SourceAssets. The returned
            assets will be copies of the loaded objects, with the prefix prepended.

    Returns:
        Sequence[Union[AssetsDefinition, SourceAsset]]:
            A list containing assets and source assets defined in the given modules.
    """

    def _asset_filter(dagster_object) -> bool:
        if isinstance(dagster_object, AssetsDefinition):
            # We don't load asset checks with asset module loaders.
            return not has_only_asset_checks(dagster_object)
        return True

    types_to_load: tuple[type] = (
        get_args(AssetLoaderTypes)
        if include_specs
        else tuple(t for t in get_args(AssetLoaderTypes) if t != AssetSpec)
    )
    return cast(
        Sequence[AssetLoaderTypes],
        ModuleScopedDagsterDefs.from_modules(modules, types_to_load=types_to_load)
        .get_object_list()
        .with_attributes(
            key_prefix=check_opt_coercible_to_asset_key_prefix_param(key_prefix, "key_prefix"),
            source_key_prefix=check_opt_coercible_to_asset_key_prefix_param(
                source_key_prefix, "source_key_prefix"
            ),
            group_name=check.opt_str_param(group_name, "group_name"),
            freshness_policy=check.opt_inst_param(
                freshness_policy, "freshness_policy", FreshnessPolicy
            ),
            automation_condition=resolve_automation_condition(
                automation_condition, auto_materialize_policy
            ),
            backfill_policy=check.opt_inst_param(
                backfill_policy, "backfill_policy", BackfillPolicy
            ),
        )
        .get_objects(_asset_filter),
    )


def load_assets_from_current_module(
    group_name: Optional[str] = None,
    key_prefix: Optional[CoercibleToAssetKeyPrefix] = None,
    *,
    freshness_policy: Optional[FreshnessPolicy] = None,
    auto_materialize_policy: Optional[AutoMaterializePolicy] = None,
    automation_condition: Optional[AutomationCondition] = None,
    backfill_policy: Optional[BackfillPolicy] = None,
    source_key_prefix: Optional[CoercibleToAssetKeyPrefix] = None,
    include_specs: bool = False,
) -> Sequence[Union[AssetsDefinition, SourceAsset, CacheableAssetsDefinition, AssetSpec]]:
    """Constructs a list of assets, source assets, and cacheable assets from the module where
    this function is called.

    Args:
        group_name (Optional[str]):
            Group name to apply to the loaded assets. The returned assets will be copies of the
            loaded objects, with the group name added.
        key_prefix (Optional[Union[str, Sequence[str]]]):
            Prefix to prepend to the keys of the loaded assets. The returned assets will be copies
            of the loaded objects, with the prefix prepended.
        freshness_policy (Optional[FreshnessPolicy]): FreshnessPolicy to apply to all the loaded
            assets.
        automation_condition (Optional[AutomationCondition]): AutomationCondition to apply
            to all the loaded assets.
        backfill_policy (Optional[AutoMaterializePolicy]): BackfillPolicy to apply to all the loaded assets.
        source_key_prefix (bool): Prefix to prepend to the keys of loaded SourceAssets. The returned
            assets will be copies of the loaded objects, with the prefix prepended.

    Returns:
        Sequence[Union[AssetsDefinition, SourceAsset, CachableAssetsDefinition]]:
            A list containing assets, source assets, and cacheable assets defined in the module.
    """
    caller = inspect.stack()[1]
    module = inspect.getmodule(caller[0])
    if module is None:
        check.failed("Could not find a module for the caller")

    return load_assets_from_modules(
        [module],
        group_name=group_name,
        key_prefix=key_prefix,
        freshness_policy=freshness_policy,
        automation_condition=resolve_automation_condition(
            automation_condition, auto_materialize_policy
        ),
        backfill_policy=backfill_policy,
        source_key_prefix=source_key_prefix,
        include_specs=include_specs,
    )


def load_assets_from_package_module(
    package_module: ModuleType,
    group_name: Optional[str] = None,
    key_prefix: Optional[CoercibleToAssetKeyPrefix] = None,
    *,
    freshness_policy: Optional[FreshnessPolicy] = None,
    auto_materialize_policy: Optional[AutoMaterializePolicy] = None,
    automation_condition: Optional[AutomationCondition] = None,
    backfill_policy: Optional[BackfillPolicy] = None,
    source_key_prefix: Optional[CoercibleToAssetKeyPrefix] = None,
    include_specs: bool = False,
) -> Sequence[Union[AssetsDefinition, SourceAsset, CacheableAssetsDefinition, AssetSpec]]:
    """Constructs a list of assets and source assets that includes all asset
    definitions, source assets, and cacheable assets in all sub-modules of the given package module.

    A package module is the result of importing a package.

    Args:
        package_module (ModuleType): The package module to looks for assets inside.
        group_name (Optional[str]):
            Group name to apply to the loaded assets. The returned assets will be copies of the
            loaded objects, with the group name added.
        key_prefix (Optional[Union[str, Sequence[str]]]):
            Prefix to prepend to the keys of the loaded assets. The returned assets will be copies
            of the loaded objects, with the prefix prepended.
        freshness_policy (Optional[FreshnessPolicy]): FreshnessPolicy to apply to all the loaded
            assets.
        automation_condition (Optional[AutomationCondition]): AutomationCondition to apply
            to all the loaded assets.
        backfill_policy (Optional[AutoMaterializePolicy]): BackfillPolicy to apply to all the loaded assets.
        source_key_prefix (bool): Prefix to prepend to the keys of loaded SourceAssets. The returned
            assets will be copies of the loaded objects, with the prefix prepended.

    Returns:
        Sequence[Union[AssetsDefinition, SourceAsset, CacheableAssetsDefinition]]:
            A list containing assets, source assets, and cacheable assets defined in the module.
    """
    return load_assets_from_modules(
        [*find_modules_in_package(package_module)],
        group_name,
        key_prefix,
        freshness_policy=freshness_policy,
        auto_materialize_policy=auto_materialize_policy,
        automation_condition=automation_condition,
        backfill_policy=backfill_policy,
        source_key_prefix=source_key_prefix,
        include_specs=include_specs,
    )


def load_assets_from_package_name(
    package_name: str,
    group_name: Optional[str] = None,
    key_prefix: Optional[CoercibleToAssetKeyPrefix] = None,
    *,
    freshness_policy: Optional[FreshnessPolicy] = None,
    auto_materialize_policy: Optional[AutoMaterializePolicy] = None,
    backfill_policy: Optional[BackfillPolicy] = None,
    source_key_prefix: Optional[CoercibleToAssetKeyPrefix] = None,
    include_specs: bool = False,
) -> Sequence[Union[AssetsDefinition, SourceAsset, CacheableAssetsDefinition, AssetSpec]]:
    """Constructs a list of assets, source assets, and cacheable assets that includes all asset
    definitions and source assets in all sub-modules of the given package.

    Args:
        package_name (str): The name of a Python package to look for assets inside.
        group_name (Optional[str]):
            Group name to apply to the loaded assets. The returned assets will be copies of the
            loaded objects, with the group name added.
        key_prefix (Optional[Union[str, Sequence[str]]]):
            Prefix to prepend to the keys of the loaded assets. The returned assets will be copies
            of the loaded objects, with the prefix prepended.
        freshness_policy (Optional[FreshnessPolicy]): FreshnessPolicy to apply to all the loaded
            assets.
        auto_materialize_policy (Optional[AutoMaterializePolicy]): AutoMaterializePolicy to apply
            to all the loaded assets.
        backfill_policy (Optional[AutoMaterializePolicy]): BackfillPolicy to apply to all the loaded assets.
        source_key_prefix (bool): Prefix to prepend to the keys of loaded SourceAssets. The returned
            assets will be copies of the loaded objects, with the prefix prepended.

    Returns:
        Sequence[Union[AssetsDefinition, SourceAsset, CacheableAssetsDefinition]]:
            A list containing assets, source assets, and cacheable assets defined in the module.
    """
    package_module = import_module(package_name)
    return load_assets_from_package_module(
        package_module,
        group_name=group_name,
        key_prefix=key_prefix,
        freshness_policy=freshness_policy,
        auto_materialize_policy=auto_materialize_policy,
        backfill_policy=backfill_policy,
        source_key_prefix=source_key_prefix,
        include_specs=include_specs,
    )



import warnings
from collections import defaultdict
from collections.abc import Mapping
from enum import Enum
from typing import TYPE_CHECKING, Optional

from dagster import (
    Field,
    Selector,
    _check as check,
)
from dagster._core.storage.dagster_run import DagsterRun, DagsterRunStatus
from dagster._core.storage.tags import MAX_RETRIES_TAG, RETRY_ON_ASSET_OR_OP_FAILURE_TAG
from dagster._serdes.serdes import whitelist_for_serdes
from dagster._utils.tags import get_boolean_tag_value

if TYPE_CHECKING:
    from dagster._core.events import RunFailureReason
    from dagster._core.instance import DagsterInstance


def get_retries_config():
    return Field(
        Selector({"enabled": {}, "disabled": {}}),
        default_value={"enabled": {}},
        description="Whether retries are enabled or not. By default, retries are enabled.",
    )


@whitelist_for_serdes
class RetryMode(Enum):
    ENABLED = "enabled"
    DISABLED = "disabled"
    # Designed for use of inner plan execution within "orchestrator" engine such as multiprocess,
    # up_for_retry steps are not directly re-enqueued, deferring that to the engine.
    DEFERRED = "deferred"

    @staticmethod
    def from_config(config_value: Mapping[str, Mapping]) -> Optional["RetryMode"]:
        for selector in config_value.keys():
            return RetryMode(selector)
        return None

    @property
    def enabled(self) -> bool:
        return self == RetryMode.ENABLED

    @property
    def disabled(self) -> bool:
        return self == RetryMode.DISABLED

    @property
    def deferred(self) -> bool:
        return self == RetryMode.DEFERRED

    def for_inner_plan(self) -> "RetryMode":
        if self.disabled or self.deferred:
            return self
        elif self.enabled:
            return RetryMode.DEFERRED
        else:
            check.failed("Unexpected RetryMode! Expected enabled, disabled, or deferred")


class RetryState:
    def __init__(self, previous_attempts: Optional[Mapping[str, int]] = None):
        self._attempts = defaultdict(int)
        for key, val in check.opt_mapping_param(
            previous_attempts, "previous_attempts", key_type=str, value_type=int
        ).items():
            self._attempts[key] = val

    def get_attempt_count(self, key: str) -> int:
        return self._attempts[key]

    def mark_attempt(self, key: str) -> None:
        self._attempts[key] += 1

    def snapshot_attempts(self) -> Mapping[str, int]:
        return dict(self._attempts)


def auto_reexecution_should_retry_run(
    instance: "DagsterInstance", run: DagsterRun, run_failure_reason: Optional["RunFailureReason"]
):
    """Determines if a run will be retried by the automatic reexcution system.
    A run will retry if:
    - it is failed.
    - the number of max allowed retries is > 0 (max retries can be set via system setting or run tag).
    - there have not already been >= max_retries retries for the run.

    If the run failure reason was a step failure and the retry_on_asset_or_op_failure tag/system setting is set to false,
    a warning message will be logged and the run will not be retried.

    We determine how many retries have been launched for the run by looking at the size of the run group
    (the set of runs that have the same root_run_id and the run with root_run_id). Since manually launched retries are
    part of the run group, this means that if a user launches a manual retry of run A and then this function
    is called because a retry for run A launched by the auto-reexecution system failed, the manual retry will be
    counted toward max_retries.

    It is unlikely, but possible, that one "extra" retry will be launched by the automatic reexecution system
    since manual retries could be happening in parallel with automatic retries. Here is
    an illustrative example:
    - Max retries is 3
    - Run A fails
    - The automatic reexecution system launches a retry of run A (A_1), which fails
    - The automatic reexecution system launches a retry run A_1 (A_2), which fails
    - This function is executing and has fetched the run_group for run A_2: (A, A_1, A_2)
    - A user launches a manual retry of run A (A_m). The run group is now (A, A_1, A_2, A_m), but this function does
    not have the updated run group
    - Since the run group we've fetched is (A, A_1, A_2), this function will mark A_2 as `will_retry=true` and
    run `A_3` will be launched. This is the "extra" retry, since usually manual retries are counted toward max_retries, but
    in this case it was not.

    We think this is an acceptable tradeoff to make since the automatic reexecution system won't launch more than max_retries
    run itself, just that max_retries + 1 runs could be launched in total if a manual retry is timed to cause this condition (unlikely).
    """
    from dagster._core.events import RunFailureReason

    if run.status != DagsterRunStatus.FAILURE:
        return False

    retry_on_asset_or_op_failure = get_boolean_tag_value(
        run.tags.get(RETRY_ON_ASSET_OR_OP_FAILURE_TAG),
        default_value=instance.run_retries_retry_on_asset_or_op_failure,
    )
    if run_failure_reason == RunFailureReason.STEP_FAILURE and not retry_on_asset_or_op_failure:
        return False

    raw_max_retries_tag = run.tags.get(MAX_RETRIES_TAG)
    if raw_max_retries_tag is None:
        max_retries = instance.run_retries_max_retries
    else:
        try:
            max_retries = int(raw_max_retries_tag)
        except ValueError:
            warnings.warn(f"Error parsing int from tag {MAX_RETRIES_TAG}, won't retry the run.")
            return False
    if max_retries > 0:
        run_group = instance.get_run_group(run.run_id)
        if run_group is not None:
            _, run_group_iter = run_group
            # since the original run is in the run group, the number of retries launched
            # so far is len(run_group_iter) - 1
            if len(list(run_group_iter)) <= max_retries:
                return True
    return False



from typing import Any


class IHasInternalInit:
    """Marker interface which indicates that this class has an dagster_internal_init method. All classes with this interface
    are unit tested to ensure that the signature of their dagster_internal_init method matches the signature of their
    __init__ method, and that dagster_internal_init has no defaults.
    """

    @staticmethod
    def dagster_internal_init(*args: Any, **kwargs: Any) -> object:
        """This method is called by the __init__ method of subclasses of IHasInternalInit. It is not intended to be called
        directly.
        """
        raise NotImplementedError(
            "This method is called by the __init__ method of subclasses of IHasInternalInit. It is not intended to be called directly."
        )



import collections.abc
import operator
import re
from abc import ABC, abstractmethod
from collections.abc import Iterable, Sequence
from functools import reduce
from typing import AbstractSet, Optional, Union, cast  # noqa: UP035

from typing_extensions import TypeAlias, TypeGuard

import dagster._check as check
from dagster._annotations import beta_param, deprecated, public
from dagster._core.definitions.asset_check_spec import AssetCheckKey
from dagster._core.definitions.asset_graph import AssetGraph
from dagster._core.definitions.asset_key import (
    AssetKey,
    CoercibleToAssetKey,
    CoercibleToAssetKeyPrefix,
    asset_keys_from_defs_and_coercibles,
    key_prefix_from_coercible,
)
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.base_asset_graph import BaseAssetGraph, BaseAssetNode
from dagster._core.definitions.resolved_asset_deps import resolve_similar_asset_names
from dagster._core.definitions.source_asset import SourceAsset
from dagster._core.errors import DagsterInvalidSubsetError
from dagster._core.selector.subset_selector import (
    fetch_connected,
    fetch_sinks,
    fetch_sources,
    parse_clause,
)
from dagster._record import copy, record
from dagster._serdes.serdes import whitelist_for_serdes

CoercibleToAssetSelection: TypeAlias = Union[
    str,
    Sequence[str],
    Sequence[AssetKey],
    Sequence[Union["AssetsDefinition", "SourceAsset"]],
    "AssetSelection",
]


def is_coercible_to_asset_selection(
    obj: object,
) -> TypeGuard[CoercibleToAssetSelection]:
    # can coerce to (but is not already) an AssetSelection
    return isinstance(obj, str) or (
        isinstance(obj, Sequence)
        and all(isinstance(x, (str, AssetKey, AssetsDefinition, SourceAsset)) for x in obj)
    )


class AssetSelection(ABC):
    """An AssetSelection defines a query over a set of assets and asset checks, normally all that are defined in a code location.

    You can use the "|", "&", and "-" operators to create unions, intersections, and differences of selections, respectively.

    AssetSelections are typically used with :py:func:`define_asset_job`.

    By default, selecting assets will also select all of the asset checks that target those assets.

    Examples:
        .. code-block:: python

            # Select all assets in group "marketing":
            AssetSelection.groups("marketing")

            # Select all assets in group "marketing", as well as the asset with key "promotion":
            AssetSelection.groups("marketing") | AssetSelection.assets("promotion")

            # Select all assets in group "marketing" that are downstream of asset "leads":
            AssetSelection.groups("marketing") & AssetSelection.assets("leads").downstream()

            # Select a list of assets:
            AssetSelection.assets(*my_assets_list)

            # Select all assets except for those in group "marketing"
            AssetSelection.all() - AssetSelection.groups("marketing")

            # Select all assets which are materialized by the same op as "projections":
            AssetSelection.assets("projections").required_multi_asset_neighbors()

            # Select all assets in group "marketing" and exclude their asset checks:
            AssetSelection.groups("marketing") - AssetSelection.all_asset_checks()

            # Select all asset checks that target a list of assets:
            AssetSelection.checks_for_assets(*my_assets_list)

            # Select a specific asset check:
            AssetSelection.checks(my_asset_check)

    """

    @public
    @staticmethod
    @beta_param(param="include_sources")
    def all(include_sources: bool = False) -> "AllSelection":
        """Returns a selection that includes all assets and their asset checks.

        Args:
            include_sources (bool): If True, then include all external assets.
        """
        return AllSelection(include_sources=include_sources)

    @public
    @staticmethod
    def all_asset_checks() -> "AllAssetCheckSelection":
        """Returns a selection that includes all asset checks."""
        return AllAssetCheckSelection()

    @public
    @staticmethod
    def assets(
        *assets_defs: Union[AssetsDefinition, CoercibleToAssetKey],
    ) -> "KeysAssetSelection":
        """Returns a selection that includes all of the provided assets and asset checks that target
        them.

        Args:
            *assets_defs (Union[AssetsDefinition, str, Sequence[str], AssetKey]): The assets to
                select.

        Examples:
            .. code-block:: python

                AssetSelection.assets(AssetKey(["a"]))

                AssetSelection.assets("a")

                AssetSelection.assets(AssetKey(["a"]), AssetKey(["b"]))

                AssetSelection.assets("a", "b")

                @asset
                def asset1():
                    ...

                AssetSelection.assets(asset1)

                asset_key_list = [AssetKey(["a"]), AssetKey(["b"])]
                AssetSelection.assets(*asset_key_list)
        """
        return KeysAssetSelection(selected_keys=asset_keys_from_defs_and_coercibles(assets_defs))

    @public
    @staticmethod
    @deprecated(
        breaking_version="2.0",
        additional_warn_text="Use AssetSelection.assets instead.",
    )
    def keys(*asset_keys: CoercibleToAssetKey) -> "KeysAssetSelection":
        """Returns a selection that includes assets with any of the provided keys and all asset
        checks that target them.

        Deprecated: use AssetSelection.assets instead.

        Examples:
            .. code-block:: python

                AssetSelection.keys(AssetKey(["a"]))

                AssetSelection.keys("a")

                AssetSelection.keys(AssetKey(["a"]), AssetKey(["b"]))

                AssetSelection.keys("a", "b")

                asset_key_list = [AssetKey(["a"]), AssetKey(["b"])]
                AssetSelection.keys(*asset_key_list)
        """
        _asset_keys = [
            AssetKey.from_user_string(key) if isinstance(key, str) else AssetKey.from_coercible(key)
            for key in asset_keys
        ]
        return KeysAssetSelection(selected_keys=_asset_keys)

    @public
    @staticmethod
    @beta_param(param="include_sources")
    def key_prefixes(
        *key_prefixes: CoercibleToAssetKeyPrefix, include_sources: bool = False
    ) -> "KeyPrefixesAssetSelection":
        """Returns a selection that includes assets that match any of the provided key prefixes and all the asset checks that target them.

        Args:
            include_sources (bool): If True, then include external assets matching the key prefix(es)
                in the selection.

        Examples:
            .. code-block:: python

              # match any asset key where the first segment is equal to "a" or "b"
              # e.g. AssetKey(["a", "b", "c"]) would match, but AssetKey(["abc"]) would not.
              AssetSelection.key_prefixes("a", "b")

              # match any asset key where the first two segments are ["a", "b"] or ["a", "c"]
              AssetSelection.key_prefixes(["a", "b"], ["a", "c"])
        """
        _asset_key_prefixes = [key_prefix_from_coercible(key_prefix) for key_prefix in key_prefixes]
        return KeyPrefixesAssetSelection(
            selected_key_prefixes=_asset_key_prefixes,
            include_sources=include_sources,
        )

    @staticmethod
    @beta_param(param="include_sources")
    def key_substring(
        key_substring: str, include_sources: bool = False
    ) -> "KeySubstringAssetSelection":
        """Returns a selection that includes assets whose string representation contains the provided substring and all the asset checks that target it.

        Args:
            include_sources (bool): If True, then include external assets matching the substring
                in the selection.

        Examples:
            .. code-block:: python

              # match any asset key containing "bc"
              # e.g. AssetKey(["a", "bcd"]) would match, but not AssetKey(["ab", "cd"]).
              AssetSelection.key_substring("bc")

              # match any asset key containing "b/c"
              # e.g. AssetKey(["ab", "cd"]) would match.
              AssetSelection.key_substring("b/c")
        """
        return KeySubstringAssetSelection(
            selected_key_substring=key_substring, include_sources=include_sources
        )

    @public
    @staticmethod
    @beta_param(param="include_sources")
    def groups(*group_strs, include_sources: bool = False) -> "GroupsAssetSelection":
        """Returns a selection that includes materializable assets that belong to any of the
        provided groups and all the asset checks that target them.

        Args:
            include_sources (bool): If True, then include external assets matching the group in the
                selection.
        """
        check.tuple_param(group_strs, "group_strs", of_type=str)
        return GroupsAssetSelection(selected_groups=group_strs, include_sources=include_sources)

    @public
    @staticmethod
    @beta_param(param="include_sources")
    def tag(key: str, value: str, include_sources: bool = False) -> "AssetSelection":
        """Returns a selection that includes materializable assets that have the provided tag, and
        all the asset checks that target them.


        Args:
            include_sources (bool): If True, then include external assets matching the group in the
                selection.
        """
        return TagAssetSelection(key=key, value=value, include_sources=include_sources)

    @staticmethod
    @beta_param(param="include_sources")
    def tag_string(string: str, include_sources: bool = False) -> "AssetSelection":
        """Returns a selection that includes materializable assets that have the provided tag, and
        all the asset checks that target them.


        Args:
            include_sources (bool): If True, then include external assets matching the group in the
                selection.
        """
        split_by_equals_segments = string.split("=")
        if len(split_by_equals_segments) == 1:
            return TagAssetSelection(key=string, value="", include_sources=include_sources)
        elif len(split_by_equals_segments) == 2:
            key, value = split_by_equals_segments
            return TagAssetSelection(key=key, value=value, include_sources=include_sources)
        else:
            check.failed(f"Invalid tag selection string: {string}. Must have no more than one '='.")

    @staticmethod
    def owner(owner: str) -> "AssetSelection":
        """Returns a selection that includes assets that have the provided owner, and all the
        asset checks that target them.

        Args:
            owner (str): The owner to select.
        """
        return OwnerAssetSelection(selected_owner=owner)

    @public
    @staticmethod
    def checks_for_assets(
        *assets_defs: Union[AssetsDefinition, CoercibleToAssetKey],
    ) -> "AssetChecksForAssetKeysSelection":
        """Returns a selection with the asset checks that target the provided assets.

        Args:
            *assets_defs (Union[AssetsDefinition, str, Sequence[str], AssetKey]): The assets to
                select checks for.
        """
        return AssetChecksForAssetKeysSelection(
            selected_asset_keys=asset_keys_from_defs_and_coercibles(assets_defs)
        )

    @public
    @staticmethod
    def checks(
        *assets_defs_or_check_keys: Union[AssetsDefinition, AssetCheckKey],
    ) -> "AssetCheckKeysSelection":
        """Returns a selection that includes all of the provided asset checks or check keys."""
        assets_defs = [ad for ad in assets_defs_or_check_keys if isinstance(ad, AssetsDefinition)]
        check_keys = [key for key in assets_defs_or_check_keys if isinstance(key, AssetCheckKey)]
        return AssetCheckKeysSelection(
            selected_asset_check_keys=[
                *(key for ad in assets_defs for key in ad.check_keys),
                *check_keys,
            ]
        )

    @public
    def downstream(
        self, depth: Optional[int] = None, include_self: bool = True
    ) -> "DownstreamAssetSelection":
        """Returns a selection that includes all assets that are downstream of any of the assets in
        this selection, selecting the assets in this selection by default. Includes the asset checks targeting the returned assets. Iterates through each
        asset in this selection and returns the union of all downstream assets.

        depth (Optional[int]): If provided, then only include assets to the given depth. A depth
            of 2 means all assets that are children or grandchildren of the assets in this
            selection.
        include_self (bool): If True, then include the assets in this selection in the result.
            If the include_self flag is False, return each downstream asset that is not part of the
            original selection. By default, set to True.
        """
        check.opt_int_param(depth, "depth")
        check.opt_bool_param(include_self, "include_self")
        return DownstreamAssetSelection(child=self, depth=depth, include_self=include_self)

    @public
    def upstream(
        self, depth: Optional[int] = None, include_self: bool = True
    ) -> "UpstreamAssetSelection":
        """Returns a selection that includes all materializable assets that are upstream of any of
        the assets in this selection, selecting the assets in this selection by default. Includes
        the asset checks targeting the returned assets. Iterates through each asset in this
        selection and returns the union of all upstream assets.

        Because mixed selections of external and materializable assets are currently not supported,
        keys corresponding to external assets will not be included as upstream of regular assets.

        Args:
            depth (Optional[int]): If provided, then only include assets to the given depth. A depth
                of 2 means all assets that are parents or grandparents of the assets in this
                selection.
            include_self (bool): If True, then include the assets in this selection in the result.
                If the include_self flag is False, return each upstream asset that is not part of the
                original selection. By default, set to True.
        """
        check.opt_int_param(depth, "depth")
        check.opt_bool_param(include_self, "include_self")
        return UpstreamAssetSelection(child=self, depth=depth, include_self=include_self)

    @public
    def sinks(self) -> "SinksAssetSelection":
        """Given an asset selection, returns a new asset selection that contains all of the sink
        assets within the original asset selection. Includes the asset checks targeting the returned assets.

        A sink asset is an asset that has no downstream dependencies within the asset selection.
        The sink asset can have downstream dependencies outside of the asset selection.
        """
        return SinksAssetSelection(child=self)

    @public
    def required_multi_asset_neighbors(self) -> "RequiredNeighborsAssetSelection":
        """Given an asset selection in which some assets are output from a multi-asset compute op
        which cannot be subset, returns a new asset selection that contains all of the assets
        required to execute the original asset selection. Includes the asset checks targeting the returned assets.
        """
        return RequiredNeighborsAssetSelection(child=self)

    @public
    def roots(self) -> "RootsAssetSelection":
        """Given an asset selection, returns a new asset selection that contains all of the root
        assets within the original asset selection. Includes the asset checks targeting the returned assets.

        A root asset is an asset that has no upstream dependencies within the asset selection.
        The root asset can have downstream dependencies outside of the asset selection.

        Because mixed selections of external and materializable assets are currently not supported,
        keys corresponding to external assets will not be included as roots. To select external assets,
        use the `upstream_source_assets` method.
        """
        return RootsAssetSelection(child=self)

    @public
    def materializable(self) -> "MaterializableAssetSelection":
        """Given an asset selection, returns a new asset selection that contains all of the assets
        that are materializable. Removes any assets which are not materializable.
        """
        return MaterializableAssetSelection(child=self)

    @public
    @deprecated(breaking_version="2.0", additional_warn_text="Use AssetSelection.roots instead.")
    def sources(self) -> "RootsAssetSelection":
        """Given an asset selection, returns a new asset selection that contains all of the root
        assets within the original asset selection. Includes the asset checks targeting the returned assets.

        A root asset is a materializable asset that has no upstream dependencies within the asset
        selection. The root asset can have downstream dependencies outside of the asset selection.

        Because mixed selections of external and materializable assets are currently not supported,
        keys corresponding to external assets will not be included as roots. To select external assets,
        use the `upstream_source_assets` method.
        """
        return self.roots()

    @public
    def upstream_source_assets(self) -> "ParentSourcesAssetSelection":
        """Given an asset selection, returns a new asset selection that contains all of the external
        assets that are parents of assets in the original selection. Includes the asset checks
        targeting the returned assets.
        """
        return ParentSourcesAssetSelection(child=self)

    @public
    def without_checks(self) -> "AssetSelection":
        """Removes all asset checks in the selection."""
        return self - AssetSelection.all_asset_checks()

    def __or__(self, other: "AssetSelection") -> "OrAssetSelection":
        check.inst_param(other, "other", AssetSelection)

        operands = []
        for selection in (self, other):
            if isinstance(selection, OrAssetSelection):
                operands.extend(selection.operands)
            else:
                operands.append(selection)

        return OrAssetSelection(operands=operands)

    def __and__(self, other: "AssetSelection") -> "AndAssetSelection":
        check.inst_param(other, "other", AssetSelection)

        operands = []
        for selection in (self, other):
            if isinstance(selection, AndAssetSelection):
                operands.extend(selection.operands)
            else:
                operands.append(selection)

        return AndAssetSelection(operands=operands)

    def __bool__(self):
        # Ensure that even if a subclass is a NamedTuple with no fields, it is still truthy
        return True

    def __sub__(self, other: "AssetSelection") -> "SubtractAssetSelection":
        check.inst_param(other, "other", AssetSelection)
        return SubtractAssetSelection(left=self, right=other)

    def resolve(
        self,
        all_assets: Union[Iterable[Union[AssetsDefinition, SourceAsset]], BaseAssetGraph],
        allow_missing: bool = False,
    ) -> AbstractSet[AssetKey]:
        """Returns the set of asset keys in all_assets that match this selection.

        Args:
            allow_missing (bool): If False, will raise an error if any of the leaf selections in the
                asset selection target entities that don't exist in the set of provided assets.
        """
        if isinstance(all_assets, BaseAssetGraph):
            asset_graph = all_assets
        else:
            check.iterable_param(all_assets, "all_assets", (AssetsDefinition, SourceAsset))
            asset_graph = AssetGraph.from_assets(all_assets)

        return self.resolve_inner(asset_graph, allow_missing=allow_missing)

    @abstractmethod
    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        raise NotImplementedError()

    def resolve_checks(
        self, asset_graph: BaseAssetGraph, allow_missing: bool = False
    ) -> AbstractSet[AssetCheckKey]:
        """We don't need this method currently, but it makes things consistent with resolve_inner. Currently
        we don't store checks in the RemoteAssetGraph, so we only support AssetGraph.
        """
        return self.resolve_checks_inner(asset_graph, allow_missing=allow_missing)

    def resolve_checks_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetCheckKey]:
        """By default, resolve to checks that target the selected assets. This is overriden for particular selections."""
        asset_keys = self.resolve(asset_graph)
        return {handle for handle in asset_graph.asset_check_keys if handle.asset_key in asset_keys}

    @classmethod
    @beta_param(param="include_sources")
    def from_string(cls, string: str, include_sources=False) -> "AssetSelection":
        from dagster._core.definitions.antlr_asset_selection.antlr_asset_selection import (
            AntlrAssetSelectionParser,
        )

        try:
            return AntlrAssetSelectionParser(string, include_sources).asset_selection
        except:
            pass
        if string == "*":
            return cls.all()

        parts = parse_clause(string)
        if parts is not None:
            key_selection = cls.assets(parts.item_name)
            if parts.up_depth and parts.down_depth:
                selection = key_selection.upstream(parts.up_depth) | key_selection.downstream(
                    parts.down_depth
                )
            elif parts.up_depth:
                selection = key_selection.upstream(parts.up_depth)
            elif parts.down_depth:
                selection = key_selection.downstream(parts.down_depth)
            else:
                selection = key_selection
            return selection

        elif string.startswith("tag:"):
            tag_str = string[len("tag:") :]
            return cls.tag_string(tag_str)

        check.failed(f"Invalid selection string: {string}")

    @classmethod
    def from_coercible(cls, selection: CoercibleToAssetSelection) -> "AssetSelection":
        if isinstance(selection, str):
            return cls.from_string(selection)
        elif isinstance(selection, AssetSelection):
            return selection
        elif isinstance(selection, collections.abc.Sequence) and all(
            isinstance(el, str) for el in selection
        ):
            return reduce(operator.or_, [cls.from_string(cast(str, s)) for s in selection])
        elif isinstance(selection, collections.abc.Sequence) and all(
            isinstance(el, (AssetsDefinition, SourceAsset)) for el in selection
        ):
            return AssetSelection.assets(
                *(
                    key
                    for el in selection
                    for key in (
                        el.keys if isinstance(el, AssetsDefinition) else [cast(SourceAsset, el).key]
                    )
                )
            )
        elif isinstance(selection, collections.abc.Sequence) and all(
            isinstance(el, AssetKey) for el in selection
        ):
            return cls.assets(*cast(Sequence[AssetKey], selection))
        else:
            check.failed(
                "selection argument must be one of str, Sequence[str], Sequence[AssetKey],"
                " Sequence[AssetsDefinition], Sequence[SourceAsset], AssetSelection. Was"
                f" {type(selection)}."
            )

    def to_serializable_asset_selection(self, asset_graph: BaseAssetGraph) -> "AssetSelection":
        return KeysAssetSelection(selected_keys=list(self.resolve(asset_graph)))

    def needs_parentheses_when_operand(self) -> bool:
        """When generating a string representation of an asset selection and this asset selection
        is an operand in a larger expression, whether it needs to be surrounded by parentheses.
        """
        return False

    def operand_to_selection_str(self) -> str:
        """Returns a string representation of the selection when it is a child of a boolean expression,
        for example, in an `AndAssetSelection` or `OrAssetSelection`. The main difference from `to_selection_str`
        is that this method may include additional parentheses around the selection to ensure that the
        expression is parsed correctly.
        """
        return (
            f"({self.to_selection_str()})"
            if self.needs_parentheses_when_operand()
            else self.to_selection_str()
        )

    def to_selection_str(self) -> str:
        """Returns an Antlr string representation of the selection that can be parsed by `from_string`."""
        raise NotImplementedError(
            f"{self.__class__.__name__} does not support conversion to a string."
        )

    def __str__(self) -> str:
        # Attempt to use the to-Antlr-selection-string method if it's implemented,
        # otherwise fall back to the default Python string representation
        try:
            return self.to_selection_str()
        except NotImplementedError:
            return super().__str__()


@whitelist_for_serdes
@record
class AllSelection(AssetSelection):
    include_sources: Optional[bool] = None

    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        return (
            asset_graph.get_all_asset_keys()
            if self.include_sources
            else asset_graph.materializable_asset_keys
        )

    def to_serializable_asset_selection(self, asset_graph: BaseAssetGraph) -> "AssetSelection":
        return self

    def to_selection_str(self) -> str:
        return "*"


@whitelist_for_serdes
@record
class AllAssetCheckSelection(AssetSelection):
    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        return set()

    def resolve_checks_inner(
        self, asset_graph: AssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetCheckKey]:
        return asset_graph.asset_check_keys

    def to_serializable_asset_selection(self, asset_graph: BaseAssetGraph) -> "AssetSelection":
        return self


@whitelist_for_serdes
@record
class AssetChecksForAssetKeysSelection(AssetSelection):
    selected_asset_keys: Sequence[AssetKey]

    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        return set()

    def resolve_checks_inner(
        self, asset_graph: AssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetCheckKey]:
        return {
            handle
            for handle in asset_graph.asset_check_keys
            if handle.asset_key in self.selected_asset_keys
        }

    def to_serializable_asset_selection(self, asset_graph: BaseAssetGraph) -> "AssetSelection":
        return self


@whitelist_for_serdes
@record
class AssetCheckKeysSelection(AssetSelection):
    selected_asset_check_keys: Sequence[AssetCheckKey]

    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        return set()

    def resolve_checks_inner(
        self, asset_graph: AssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetCheckKey]:
        specified_keys = set(self.selected_asset_check_keys)
        missing_keys = {key for key in specified_keys if key not in asset_graph.asset_check_keys}

        if not allow_missing and missing_keys:
            raise DagsterInvalidSubsetError(
                f"AssetCheckKey(s) {[k.to_user_string() for k in missing_keys]} were selected, but "
                "no definitions supply these keys. Make sure all keys are spelled "
                "correctly, and all definitions are correctly added to the "
                f"`Definitions`."
            )
        return specified_keys & asset_graph.asset_check_keys

    def to_serializable_asset_selection(self, asset_graph: BaseAssetGraph) -> "AssetSelection":
        return self


@record
class OperandListAssetSelection(AssetSelection):
    """Superclass for classes like `AndAssetSelection` and `OrAssetSelection` that operate on
    a list of sub-AssetSelections.
    """

    operands: Sequence[AssetSelection]

    def to_serializable_asset_selection(self, asset_graph: BaseAssetGraph) -> "AssetSelection":
        return copy(
            self,
            operands=[
                operand.to_serializable_asset_selection(asset_graph) for operand in self.operands
            ],
        )

    def __eq__(self, other):
        if not isinstance(other, OperandListAssetSelection):
            return False

        num_operands = len(self.operands)
        return len(other.operands) == num_operands and all(
            self.operands[i] == other.operands[i] for i in range(num_operands)
        )

    def needs_parentheses_when_operand(self) -> bool:
        return True


@whitelist_for_serdes
class AndAssetSelection(OperandListAssetSelection):
    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        return reduce(
            operator.and_,
            (
                selection.resolve_inner(asset_graph, allow_missing=allow_missing)
                for selection in self.operands
            ),
        )

    def resolve_checks_inner(
        self, asset_graph: AssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetCheckKey]:
        return reduce(
            operator.and_,
            (
                selection.resolve_checks_inner(asset_graph, allow_missing=allow_missing)
                for selection in self.operands
            ),
        )

    def to_selection_str(self) -> str:
        return " and ".join(f"{operand.operand_to_selection_str()}" for operand in self.operands)


@whitelist_for_serdes
class OrAssetSelection(OperandListAssetSelection):
    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        return reduce(
            operator.or_,
            (
                selection.resolve_inner(asset_graph, allow_missing=allow_missing)
                for selection in self.operands
            ),
        )

    def resolve_checks_inner(
        self, asset_graph: AssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetCheckKey]:
        return reduce(
            operator.or_,
            (
                selection.resolve_checks_inner(asset_graph, allow_missing=allow_missing)
                for selection in self.operands
            ),
        )

    def to_selection_str(self) -> str:
        return " or ".join(f"{operand.operand_to_selection_str()}" for operand in self.operands)


@whitelist_for_serdes
@record
class SubtractAssetSelection(AssetSelection):
    left: AssetSelection
    right: AssetSelection

    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        return self.left.resolve_inner(
            asset_graph, allow_missing=allow_missing
        ) - self.right.resolve_inner(asset_graph, allow_missing=allow_missing)

    def resolve_checks_inner(
        self, asset_graph: AssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetCheckKey]:
        return self.left.resolve_checks_inner(
            asset_graph, allow_missing=allow_missing
        ) - self.right.resolve_checks_inner(asset_graph, allow_missing=allow_missing)

    def to_serializable_asset_selection(self, asset_graph: BaseAssetGraph) -> "AssetSelection":
        return copy(
            self,
            left=self.left.to_serializable_asset_selection(asset_graph),
            right=self.right.to_serializable_asset_selection(asset_graph),
        )

    def needs_parentheses_when_operand(self) -> bool:
        return True

    def to_selection_str(self) -> str:
        if isinstance(self.left, AllSelection):
            return f"not {self.right.to_selection_str()}"
        return f"{self.left.operand_to_selection_str()} and not {self.right.operand_to_selection_str()}"


@record
class ChainedAssetSelection(AssetSelection):
    """Superclass for AssetSelection classes that contain a single child AssetSelection and are
    resolved by applying some operation to the result of resolving the child selection.
    """

    child: AssetSelection

    def to_serializable_asset_selection(self, asset_graph: BaseAssetGraph) -> "AssetSelection":
        return copy(self, child=self.child.to_serializable_asset_selection(asset_graph))


@whitelist_for_serdes
class SinksAssetSelection(ChainedAssetSelection):
    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        selection = self.child.resolve_inner(asset_graph, allow_missing=allow_missing)
        return fetch_sinks(asset_graph.asset_dep_graph, selection)

    def to_selection_str(self) -> str:
        return f"sinks({self.child.to_selection_str()})"


@whitelist_for_serdes
class RequiredNeighborsAssetSelection(ChainedAssetSelection):
    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        selection = self.child.resolve_inner(asset_graph, allow_missing=allow_missing)
        output = set(selection)
        for asset_key in selection:
            output.update(asset_graph.get(asset_key).execution_set_asset_keys)
        return output


@whitelist_for_serdes
class RootsAssetSelection(ChainedAssetSelection):
    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        selection = self.child.resolve_inner(asset_graph, allow_missing=allow_missing)
        return fetch_sources(asset_graph, selection)

    def to_selection_str(self) -> str:
        return f"roots({self.child.to_selection_str()})"


@whitelist_for_serdes
class MaterializableAssetSelection(ChainedAssetSelection):
    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        return {
            asset_key
            for asset_key in self.child.resolve_inner(asset_graph, allow_missing=allow_missing)
            if cast(BaseAssetNode, asset_graph.get(asset_key)).is_materializable
        }


@whitelist_for_serdes
@record
class DownstreamAssetSelection(ChainedAssetSelection):
    depth: Optional[int]
    include_self: bool

    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        selection = self.child.resolve_inner(asset_graph, allow_missing=allow_missing)
        return operator.sub(
            reduce(
                operator.or_,
                [
                    {asset_key}
                    | fetch_connected(
                        item=asset_key,
                        graph=asset_graph.asset_dep_graph,
                        direction="downstream",
                        depth=self.depth,
                    )
                    for asset_key in selection
                ],
            ),
            selection if not self.include_self else set(),
        )

    def to_selection_str(self) -> str:
        if self.depth is None:
            base = f"{self.child.operand_to_selection_str()}+"
        elif self.depth == 0:
            base = self.child.operand_to_selection_str()
        else:
            base = f"{self.child.operand_to_selection_str()}{'+'}{self.depth}"

        if self.include_self:
            return base
        else:
            return f"{base} and not {self.child.operand_to_selection_str()}"


@whitelist_for_serdes
@record
class GroupsAssetSelection(AssetSelection):
    selected_groups: Sequence[str]
    include_sources: bool

    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        base_set = (
            asset_graph.get_all_asset_keys()
            if self.include_sources
            else asset_graph.materializable_asset_keys
        )
        return {
            key
            for group in self.selected_groups
            for key in asset_graph.asset_keys_for_group(group)
            if key in base_set
        }

    def to_serializable_asset_selection(self, asset_graph: BaseAssetGraph) -> "AssetSelection":
        return self

    def needs_parentheses_when_operand(self) -> bool:
        return len(self.selected_groups) > 1

    def to_selection_str(self) -> str:
        if len(self.selected_groups) == 1:
            return f'group:"{self.selected_groups[0]}"'
        else:
            return " or ".join(f'group:"{group}"' for group in self.selected_groups)


@whitelist_for_serdes
@record
class TagAssetSelection(AssetSelection):
    key: str
    value: str
    include_sources: bool

    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        base_set = (
            asset_graph.get_all_asset_keys()
            if self.include_sources
            else asset_graph.materializable_asset_keys
        )

        return {key for key in base_set if asset_graph.get(key).tags.get(self.key) == self.value}

    def to_selection_str(self) -> str:
        if self.value:
            return f'tag:"{self.key}"="{self.value}"'
        else:
            return f'tag:"{self.key}"'


@whitelist_for_serdes
@record
class OwnerAssetSelection(AssetSelection):
    selected_owner: str

    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        return {
            key
            for key in asset_graph.get_all_asset_keys()
            if self.selected_owner in asset_graph.get(key).owners
        }

    def to_selection_str(self) -> str:
        return f'owner:"{self.selected_owner}"'


@whitelist_for_serdes
@record
class CodeLocationAssetSelection(AssetSelection):
    """Used to represent a UI asset selection by code location. This should not be resolved against
    an in-process asset graph.
    """

    selected_code_location: str

    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        """This should not be invoked in user code."""
        raise NotImplementedError

    def to_selection_str(self) -> str:
        return f'code_location:"{self.selected_code_location}"'


@whitelist_for_serdes
@record
class ColumnAssetSelection(AssetSelection):
    """Used to represent a UI asset selection by column. This should not be resolved against
    an in-process asset graph.
    """

    selected_column: str

    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        """This should not be invoked in user code."""
        raise NotImplementedError

    def to_selection_str(self) -> str:
        return f'column:"{self.selected_column}"'


@whitelist_for_serdes
@record
class TableNameAssetSelection(AssetSelection):
    """Used to represent a UI asset selection by table name. This should not be resolved against
    an in-process asset graph.
    """

    selected_table_name: str

    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        """This should not be invoked in user code."""
        raise NotImplementedError

    def to_selection_str(self) -> str:
        return f'table_name:"{self.selected_table_name}"'


@whitelist_for_serdes
@record
class ColumnTagAssetSelection(AssetSelection):
    """Used to represent a UI asset selection by column tag. This should not be resolved against
    an in-process asset graph.
    """

    key: str
    value: str

    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        """This should not be invoked in user code."""
        raise NotImplementedError

    def to_selection_str(self) -> str:
        if self.value:
            return f'column_tag:"{self.key}"="{self.value}"'
        else:
            return f'column_tag:"{self.key}"'


@whitelist_for_serdes
@record
class ChangedInBranchAssetSelection(AssetSelection):
    """Used to represent a UI asset selection by changed in branch metadata. This should not be resolved against
    an in-process asset graph.
    """

    selected_changed_in_branch: str

    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        """This should not be invoked in user code."""
        raise NotImplementedError

    def to_selection_str(self) -> str:
        return f'changed_in_branch:"{self.selected_changed_in_branch}"'


@whitelist_for_serdes
@record
class KeysAssetSelection(AssetSelection):
    selected_keys: Sequence[AssetKey]

    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        specified_keys = set(self.selected_keys)
        missing_keys = {key for key in specified_keys if not asset_graph.has(key)}

        if not allow_missing:
            # Arbitrary limit to avoid huge error messages
            keys_to_suggest = list(missing_keys)[:4]
            suggestions = ""
            for invalid_key in keys_to_suggest:
                similar_names = resolve_similar_asset_names(
                    invalid_key, asset_graph.get_all_asset_keys()
                )
                if similar_names:
                    # Arbitrarily limit to 10 similar names to avoid a huge error message
                    subset_similar_names = similar_names[:10]
                    similar_to_string = ", ".join(
                        similar.to_string() for similar in subset_similar_names
                    )
                    suggestions += (
                        f"\n\nFor selected asset {invalid_key.to_string()}, did you mean one of "
                        f"the following?\n\t{similar_to_string}"
                    )

            if missing_keys:
                raise DagsterInvalidSubsetError(
                    f"AssetKey(s) {[k.to_user_string() for k in missing_keys]} were selected, but "
                    "no AssetsDefinition objects supply these keys. Make sure all keys are spelled "
                    "correctly, and all AssetsDefinitions are correctly added to the "
                    f"`Definitions`.{suggestions}"
                )

        return specified_keys - missing_keys

    def to_serializable_asset_selection(self, asset_graph: BaseAssetGraph) -> "AssetSelection":
        return self

    def needs_parentheses_when_operand(self) -> bool:
        return len(self.selected_keys) > 1

    def to_selection_str(self) -> str:
        return " or ".join(f'key:"{x.to_user_string()}"' for x in self.selected_keys)


@whitelist_for_serdes
@record
class KeyPrefixesAssetSelection(AssetSelection):
    selected_key_prefixes: Sequence[Sequence[str]]
    include_sources: bool

    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        base_set = (
            asset_graph.get_all_asset_keys()
            if self.include_sources
            else asset_graph.materializable_asset_keys
        )
        return {
            key
            for key in base_set
            if any(key.has_prefix(prefix) for prefix in self.selected_key_prefixes)
        }

    def to_serializable_asset_selection(self, asset_graph: BaseAssetGraph) -> "AssetSelection":
        return self


@whitelist_for_serdes
@record
class KeySubstringAssetSelection(AssetSelection):
    selected_key_substring: str
    include_sources: bool

    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        base_set = (
            asset_graph.get_all_asset_keys()
            if self.include_sources
            else asset_graph.materializable_asset_keys
        )
        return {key for key in base_set if self.selected_key_substring in key.to_user_string()}

    def to_serializable_asset_selection(self, asset_graph: BaseAssetGraph) -> "AssetSelection":
        return self

    def to_selection_str(self) -> str:
        return f'key_substring:"{self.selected_key_substring}"'


@whitelist_for_serdes
@record
class KeyWildCardAssetSelection(AssetSelection):
    selected_key_wildcard: str

    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        regex = re.compile("^" + re.escape(self.selected_key_wildcard).replace("\\*", ".*") + "$")
        return {
            key for key in asset_graph.get_all_asset_keys() if regex.match(key.to_user_string())
        }

    def to_selection_str(self) -> str:
        return f'key:"{self.selected_key_wildcard}"'


def _fetch_all_upstream(
    selection: AbstractSet[AssetKey],
    asset_graph: BaseAssetGraph,
    depth: Optional[int] = None,
    include_self: bool = True,
) -> AbstractSet[AssetKey]:
    return operator.sub(
        reduce(
            operator.or_,
            [
                {asset_key}
                | fetch_connected(
                    item=asset_key,
                    graph=asset_graph.asset_dep_graph,
                    direction="upstream",
                    depth=depth,
                )
                for asset_key in selection
            ],
            set(),
        ),
        selection if not include_self else set(),
    )


@whitelist_for_serdes
@record
class UpstreamAssetSelection(ChainedAssetSelection):
    depth: Optional[int]
    include_self: bool

    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        selection = self.child.resolve_inner(asset_graph, allow_missing=allow_missing)
        if len(selection) == 0:
            return selection
        all_upstream = _fetch_all_upstream(selection, asset_graph, self.depth, self.include_self)
        return {key for key in all_upstream if key in asset_graph.materializable_asset_keys}

    def to_selection_str(self) -> str:
        if self.depth is None:
            base = f"+{self.child.operand_to_selection_str()}"
        elif self.depth == 0:
            base = self.child.operand_to_selection_str()
        else:
            base = f"{self.depth}{'+'}{self.child.operand_to_selection_str()}"

        if self.include_self:
            return base
        else:
            return f"{base} and not {self.child.operand_to_selection_str()}"


@whitelist_for_serdes
class ParentSourcesAssetSelection(ChainedAssetSelection):
    def resolve_inner(
        self, asset_graph: BaseAssetGraph, allow_missing: bool
    ) -> AbstractSet[AssetKey]:
        selection = self.child.resolve_inner(asset_graph, allow_missing=allow_missing)
        if len(selection) == 0:
            return selection
        all_upstream = _fetch_all_upstream(selection, asset_graph)
        return {key for key in all_upstream if key in asset_graph.external_asset_keys}



"""Terms and concepts in lazy / freshness-based auto-materialize:
- data_time: see data_time.py
- effective_data_time: The data time that this asset would have if the most recent run succeeded.
    If the most recent run completed successfully / is not in progress or failed, then this is
    just the current data time of the asset.
- execution_period: The range of times in which it would be acceptable to materialize this asset,
    i.e. it`s late enough to pull in the required data time, and early enough to not go over the
    maximum lag minutes.
"""

import datetime
from collections.abc import Sequence
from typing import TYPE_CHECKING, AbstractSet, Optional  # noqa: UP035

from dagster._core.definitions.declarative_automation.legacy.valid_asset_subset import (
    ValidAssetSubset,
)
from dagster._core.definitions.events import AssetKeyPartitionKey
from dagster._core.definitions.freshness_policy import FreshnessPolicy
from dagster._core.definitions.time_window_partitions import TimeWindow
from dagster._utils.schedules import cron_string_iterator

if TYPE_CHECKING:
    from dagster._core.definitions.auto_materialize_rule_evaluation import TextRuleEvaluationData
    from dagster._core.definitions.declarative_automation.legacy.legacy_context import (
        LegacyRuleEvaluationContext,
    )
    from dagster._core.definitions.declarative_automation.serialized_objects import (
        AssetSubsetWithMetadata,
    )


def get_execution_period_for_policy(
    freshness_policy: FreshnessPolicy,
    effective_data_time: Optional[datetime.datetime],
    current_time: datetime.datetime,
) -> TimeWindow:
    if freshness_policy.cron_schedule:
        tick_iterator = cron_string_iterator(
            start_timestamp=current_time.timestamp(),
            cron_string=freshness_policy.cron_schedule,
            execution_timezone=freshness_policy.cron_schedule_timezone,
        )

        while True:
            # find the next tick that requires data after the current effective data time
            # (usually, this will be the next tick)
            tick = next(tick_iterator)
            required_data_time = tick - freshness_policy.maximum_lag_delta
            if effective_data_time is None or effective_data_time < required_data_time:
                return TimeWindow(start=required_data_time, end=tick)

    else:
        # occurs when asset is missing
        if effective_data_time is None:
            return TimeWindow(
                # require data from at most maximum_lag_delta ago
                start=current_time - freshness_policy.maximum_lag_delta,
                # this data should be available as soon as possible
                end=current_time,
            )
        return TimeWindow(
            # we don't want to execute this too frequently
            start=effective_data_time + 0.9 * freshness_policy.maximum_lag_delta,
            end=max(effective_data_time + freshness_policy.maximum_lag_delta, current_time),
        )


def get_execution_period_and_evaluation_data_for_policies(
    local_policy: Optional[FreshnessPolicy],
    policies: AbstractSet[FreshnessPolicy],
    effective_data_time: Optional[datetime.datetime],
    current_time: datetime.datetime,
) -> tuple[Optional[TimeWindow], Optional["TextRuleEvaluationData"]]:
    """Determines a range of times for which you can kick off an execution of this asset to solve
    the most pressing constraint, alongside a maximum number of additional constraints.
    """
    from dagster._core.definitions.auto_materialize_rule_evaluation import TextRuleEvaluationData

    merged_period = None
    contains_local = False
    contains_downstream = False
    for period, policy in sorted(
        (
            (get_execution_period_for_policy(policy, effective_data_time, current_time), policy)
            for policy in policies
        ),
        # sort execution periods by most pressing
        key=lambda pp: pp[0].end,
    ):
        if merged_period is None:
            merged_period = period
        elif period.start <= merged_period.end:
            merged_period = TimeWindow(
                start=max(period.start, merged_period.start),
                end=period.end,
            )
        else:
            break

        if policy == local_policy:
            contains_local = True
        else:
            contains_downstream = True

    if not contains_local and not contains_downstream:
        evaluation_data = None
    elif not contains_local:
        evaluation_data = TextRuleEvaluationData("Required by downstream asset's policy")
    elif not contains_downstream:
        evaluation_data = TextRuleEvaluationData("Required by this asset's policy")
    else:
        evaluation_data = TextRuleEvaluationData(
            "Required by this asset's policy and downstream asset's policy"
        )

    return merged_period, evaluation_data


def get_expected_data_time_for_asset_key(
    context: "LegacyRuleEvaluationContext", will_materialize: bool
) -> Optional[datetime.datetime]:
    """Returns the data time that you would expect this asset to have if you were to execute it
    on this tick.
    """
    from dagster._core.definitions.remote_asset_graph import RemoteWorkspaceAssetGraph

    asset_key = context.asset_key
    asset_graph = context.asset_graph
    current_time = context.evaluation_time

    # don't bother calculating if no downstream assets have freshness policies
    if not asset_graph.get_downstream_freshness_policies(asset_key=asset_key):
        return None
    # if asset will not be materialized, just return the current time
    elif not will_materialize:
        return context.data_time_resolver.get_current_data_time(asset_key, current_time)
    elif asset_graph.has_materializable_parents(asset_key):
        expected_data_time = None
        for parent_key in asset_graph.get(asset_key).parent_keys:
            # if the parent will be materialized on this tick, and it's not in the same repo, then
            # we must wait for this asset to be materialized
            if isinstance(
                asset_graph, RemoteWorkspaceAssetGraph
            ) and context.will_update_asset_partition(AssetKeyPartitionKey(parent_key)):
                parent_repo = asset_graph.get_repository_handle(parent_key)
                if parent_repo != asset_graph.get_repository_handle(asset_key):
                    return context.data_time_resolver.get_current_data_time(asset_key, current_time)
            # find the minimum non-None data time of your parents
            parent_expected_data_time = context.expected_data_time_mapping.get(
                parent_key
            ) or context.data_time_resolver.get_current_data_time(parent_key, current_time)
            expected_data_time = min(
                filter(None, [expected_data_time, parent_expected_data_time]),
                default=None,
            )
        return expected_data_time
    # for root assets, this would just be the current time
    else:
        return current_time


def freshness_evaluation_results_for_asset_key(
    context: "LegacyRuleEvaluationContext",
) -> tuple[ValidAssetSubset, Sequence["AssetSubsetWithMetadata"]]:
    """Returns a set of AssetKeyPartitionKeys to materialize in order to abide by the given
    FreshnessPolicies.

    Attempts to minimize the total number of asset executions.
    """
    from dagster._core.definitions.declarative_automation.serialized_objects import (
        AssetSubsetWithMetadata,
    )

    asset_key = context.asset_key
    current_time = context.evaluation_time

    if (
        not context.asset_graph.get_downstream_freshness_policies(asset_key=asset_key)
        or context.asset_graph.get(asset_key).is_partitioned
    ):
        return context.empty_subset(), []

    # figure out the current contents of this asset
    current_data_time = context.data_time_resolver.get_current_data_time(asset_key, current_time)

    # figure out the data time you would expect if you were to execute this asset on this tick
    expected_data_time = get_expected_data_time_for_asset_key(
        context=context,
        will_materialize=True,
    )

    # if executing the asset on this tick would not change its data time, then return
    if current_data_time == expected_data_time:
        return context.empty_subset(), []

    # calculate the data times you would expect after all currently-executing runs
    # were to successfully complete
    in_progress_data_time = context.data_time_resolver.get_in_progress_data_time(
        asset_key, current_time
    )

    # calculate the data times you would have expected if the most recent run succeeded
    failed_data_time = context.data_time_resolver.get_ignored_failure_data_time(
        asset_key, current_time
    )

    effective_data_time = max(
        filter(None, (current_data_time, in_progress_data_time, failed_data_time)),
        default=None,
    )

    # figure out a time period that you can execute this asset within to solve a maximum
    # number of constraints
    (
        execution_period,
        evaluation_data,
    ) = get_execution_period_and_evaluation_data_for_policies(
        local_policy=context.asset_graph.get(asset_key).freshness_policy,
        policies=context.asset_graph.get_downstream_freshness_policies(asset_key=asset_key),
        effective_data_time=effective_data_time,
        current_time=current_time,
    )

    if (
        execution_period is not None
        and execution_period.start <= current_time
        and expected_data_time is not None
        # if this is False, then executing it would still leave the asset overdue
        and expected_data_time >= execution_period.start
        and evaluation_data is not None
    ):
        all_subset = ValidAssetSubset.all(asset_key, None)
        return (
            ValidAssetSubset.all(asset_key, None),
            [AssetSubsetWithMetadata(subset=all_subset, metadata=evaluation_data.metadata)],
        )
    else:
        return context.empty_subset(), []
